

import copy
import logging
import os.path
import pprint
import re
import fnmatch

import wayround_org.utils.list
import wayround_org.utils.tag
import wayround_org.utils.text


# Difficult name examples:
DIFFICULT_NAMES = [
    'GeSHi-1.0.2-beta-1.tar.bz2',
    'Perl-Dist-Strawberry-BuildPerl-5101-2.11_10.tar.gz',
    'bind-9.9.1-P2.tar.gz',
    'boost_1_25_1.tar.bz2',
    'dahdi-linux-complete-2.1.0.3+2.1.0.2.tar.gz',
    'dhcp-4.1.2rc1.tar.gz',
    'dvd+rw-tools-5.5.4.3.4.tar.gz',
    'gtk+-3.12.0.tar.xz',
    'lynx2.8.7rel.1.tar.bz2',
    'name.tar.gz',
    'ogre_src_v1-8-1.tar.bz2',
    'openjdk-8-src-b132-03_mar_2014.zip',
    'openssl-0.9.7a.tar.gz',
    'org.apache.felix.ipojo.manipulator-1.8.4-project.tar.gz',
    'pkcs11-helper-1.05.tar.bz2',
    'qca-pkcs11-0.1-20070425.tar.bz2',
    'tcl8.4.19-src.tar.gz',
    'wmirq-0.1-source.tar.gz',
    'xc-1.tar.gz',
    'xf86-input-acecad-1.5.0.tar.bz2',
    'xf86-input-elo2300-1.1.2.tar.bz2',

    # delimiters missing between version numbers :-
    'unzip60.tar.gz',
    'zip30.tar.gz',
    'zip30c.tar.gz'
    ]
"""
Testing tarbal names
"""

ACCEPTABLE_SOURCE_NAME_EXTENSIONS = [
    '.tar.gz',
    '.tar.bz2',
    '.tar.xz',
    '.tar.lzma',
    '.zip',
    '.7z',
    '.tgz',
    '.tbz2',
    '.tbz'
    ]
"""
Acceptable source name extensions
"""

ALL_DELIMITERS = ['.', '_', '-', '~']
STATUS_DELIMITERS = ALL_DELIMITERS + ['+']


def _find_possible_chared_versions_and_singles(name_sliced, separator='.'):
    """
    From sliced package name, return all possible versions
    """

    versions = []
    logging.debug(
        "(internal1) versions delimitered by `{}': {}".format(
            separator,
            versions
            )
        )

    version_started = None
    version_ended = None

    index = -1

    for i in name_sliced:
        index += 1

        if i.isdecimal():

            if version_started is None:
                version_started = index

            version_ended = index

        else:

            if version_started is not None:
                if i != separator:
                    versions.append((version_started, version_ended + 1,))
                    version_started = None

    if version_started is not None:
        versions.append((version_started, version_ended + 1,))
        version_started = None

    logging.debug(
        "(internal2) versions delimitered by `{}': {}".format(
            separator,
            versions
            )
        )

    singles = []
    multiples = []

    for i in versions:
        if i[1] - i[0] == 1:
            singles.append(i)
        elif i[1] - i[0] > 1:
            multiples.append(i)
        else:
            raise Exception("Programming error")

    logging.debug(
        "(internal3) versions delimitered by `{}': {}".format(
            separator,
            versions
            )
        )

    return {'singles': singles, 'version': multiples}


def _find_all_versions_and_singles(name_sliced):
    """
    Find all versions using :func:`_find_possible_chared_versions_and_singles`
    function
    """
    ret = dict()
    for i in ALL_DELIMITERS:
        ret[i] = _find_possible_chared_versions_and_singles(name_sliced, i)
        logging.debug("versions delimitered by `{}': {}".format(i, ret[i]))
    return ret


def standard_version_finder(name_sliced, mute=False):
    """
    Find most possible version in sliced package name
    """

    ret = None

    possible_versions_and_singles_grouped_by_delimeter = \
        _find_all_versions_and_singles(name_sliced)

    logging.debug(
        "possible_versions_and_singles_grouped_by_delimeter: {}".format(
            repr(
                possible_versions_and_singles_grouped_by_delimeter
                )
            )
        )

    possible_versions_grouped_by_delimeter = {}
    possible_singles_grouped_by_delimeter = {}

    for i in ALL_DELIMITERS:

        possible_versions_grouped_by_delimeter[i] = \
            possible_versions_and_singles_grouped_by_delimeter[i]['version']

        possible_singles_grouped_by_delimeter[i] = \
            possible_versions_and_singles_grouped_by_delimeter[i]['singles']

    for i in ALL_DELIMITERS:

        if isinstance(ret, (tuple, int)):
            break

        l_possible_versions_grouped_by_delimeter_i = (
            len(possible_versions_grouped_by_delimeter[i])
            )

        if l_possible_versions_grouped_by_delimeter_i == 0:
            pass

        elif l_possible_versions_grouped_by_delimeter_i == 1:
            ret = possible_versions_grouped_by_delimeter[i][0]
            break
        else:

            current_delimiter_group = possible_versions_grouped_by_delimeter[i]

            maximum_length = 0

            for j in current_delimiter_group:
                l = j[1] - j[0]
                if l > maximum_length:
                    maximum_length = l

            if maximum_length == 0:
                s = "Version not found in group `{}'".format(i)
                if not mute:
                    logging.error(s)
                else:
                    logging.debug(s)
            else:

                lists_to_compare = []

                logging.debug(
                    "lists_to_compare: {}".format(repr(lists_to_compare))
                    )

                for j in current_delimiter_group:
                    l = j[1] - j[0]
                    if l == maximum_length:
                        lists_to_compare.append(j)

                l = len(lists_to_compare)
                if l == 0:
                    ret = None
                elif l == 1:
                    ret = lists_to_compare[0]
                else:

                    most_possible_version2 = lists_to_compare[0]

                    for j in lists_to_compare:
                        if j[0] < most_possible_version2[0]:
                            most_possible_version2 = j

                    logging.debug(
                        "most_possible_version2: {}".format(
                            repr(most_possible_version2)
                            )
                        )
                    ret = most_possible_version2
                    break

    if ret is None:
        for i in ALL_DELIMITERS:

            if isinstance(ret, (tuple, int)):
                break

            l_possible_singles_grouped_by_delimeter_i = (
                len(possible_singles_grouped_by_delimeter[i])
                )

            if l_possible_singles_grouped_by_delimeter_i == 0:
                pass

            elif l_possible_singles_grouped_by_delimeter_i == 1:
                ret = possible_singles_grouped_by_delimeter[i][0]
                break
            else:

                most_possible_version3 = \
                    possible_singles_grouped_by_delimeter[i][0]

                for j in possible_singles_grouped_by_delimeter[i]:
                    if j[0] < most_possible_version3[0]:
                        most_possible_version3 = j

                logging.debug(
                    "most_possible_version3: {}".format(
                        repr(most_possible_version3)
                        )
                    )
                ret = most_possible_version3
                break

    logging.debug("most_possible_version: {}".format(repr(ret)))

    return ret


def standard_version_splitter(name_sliced, most_possible_version):

    ret = dict(
        version=None,
        version_list_dirty=None,
        version_list=None,
        version_dirty=None
        )

    ret['version_list_dirty'] = \
        name_sliced[most_possible_version[0]:most_possible_version[1]]

    ret['version_list'] = \
        copy.copy(ret['version_list_dirty'])

    wayround_org.utils.list.remove_all_values(
        ret['version_list'],
        ALL_DELIMITERS
        )

    ret['version'] = '.'.join(ret['version_list'])

    ret['version_dirty'] = ''.join(ret['version_list_dirty'])

    return ret


def infozip_version_finder(name_sliced, mute=False):
    return 1, 2


def infozip_version_splitter(name_sliced, most_possible_version):

    ret = None

    if len(name_sliced) > 1:

        ret = dict(
            version=None,
            version_list_dirty=None,
            version_list=None,
            version_dirty=None
            )

        value = list(name_sliced[1])

        ret['version_list_dirty'] = value

        ret['version_list'] = \
            copy.copy(ret['version_list_dirty'])

        ret['version'] = '.'.join(ret['version_list'])

        ret['version_dirty'] = ''.join(ret['version_list_dirty'])

    return ret


def standard_version_functions_selector(filebn, subject):

    ret = None

    if not subject in ['finder', 'splitter']:
        raise ValueError("invalid `subject' value")

    if filebn.startswith('zip') or filebn.startswith('unzip'):

        if subject == 'finder':
            ret = infozip_version_finder

        elif subject == 'splitter':
            ret = infozip_version_splitter

    else:

        if subject == 'finder':
            ret = standard_version_finder

        elif subject == 'splitter':
            ret = standard_version_splitter

    return ret


def parse_tarball_name(
        filename, mute=False, acceptable_source_name_extensions=None,
        version_functions_selector=None
        ):
    """
    Parse source tarball name

    If this function succeeded, dict is returned::

        {
            'name': None,
            'groups': {
                'name'              : None,
                'extension'         : None,

                'version'           : None,
                'version_list_dirty': None,
                'version_list'      : None,
                'version_dirty'     : None,

                'status'            : None,
                'status_list_dirty' : None,
                'status_dirty'      : None,
                'status_list'       : None,
                }
            }

    .. NOTE:: ret['group']['version'] numbers are always joined with point(s)

    see standard_version_finder for reference version finder

    see standard_version_splitter for reference version finder
    """

    if not isinstance(filename, str):
        raise TypeError("filename must be str")

    if acceptable_source_name_extensions is None:
        acceptable_source_name_extensions = ACCEPTABLE_SOURCE_NAME_EXTENSIONS

    if version_functions_selector is None:
        version_functions_selector = standard_version_functions_selector

    filename = os.path.basename(filename)

    logging.debug("Parsing source file name {}".format(filename))

    ret = None

    extension = None
    for i in acceptable_source_name_extensions:
        if filename.endswith(i):
            extension = i

    if extension is None:
        s = "Wrong extension"
        if not mute:
            logging.error(s)
        else:
            logging.debug(s)
        ret = 1
    else:
        without_extension = filename[:-len(extension)]

        # TODO: copy this parser to this module
        name_sliced = wayround_org.utils.text.slice_string_to_sections(
            without_extension
            )

        version_finder = version_functions_selector(filename, 'finder')

        most_possible_version = version_finder(name_sliced, mute)

        if not isinstance(most_possible_version, tuple):
            ret = None
        else:
            ret = {
                'name': None,
                'groups': {
                    'name': None,
                    'extension': None,

                    'version': None,
                    'version_list_dirty': None,
                    'version_list': None,
                    'version_dirty': None,

                    'status': None,
                    'status_list_dirty': None,
                    'status_dirty': None,
                    'status_list': None,
                    }
                }

            ret['name'] = filename

            ret['groups']['name'] = ''.join(
                name_sliced[:most_possible_version[0]]
                ).strip(''.join(ALL_DELIMITERS))

            # version operations

            version_splitter = version_functions_selector(filename, 'splitter')
            version_splitter_res = version_splitter(
                name_sliced,
                most_possible_version
                )

            if version_splitter_res is not None:

                ret['groups'].update(version_splitter_res)

                # status operations

                ret['groups']['status_list_dirty'] = (
                    name_sliced[most_possible_version[1]:]
                    )

                ret['groups']['status_list_dirty'] = (
                    wayround_org.utils.list.list_strip(
                        ret['groups']['status_list_dirty'],
                        STATUS_DELIMITERS
                        )
                    )

                ret['groups']['status_list'] = (
                    copy.copy(ret['groups']['status_list_dirty'])
                    )

                wayround_org.utils.list.remove_all_values(
                    ret['groups']['status_list'],
                    STATUS_DELIMITERS
                    )

                ret['groups']['status_list'] = (
                    wayround_org.utils.list.list_strip(
                        ret['groups']['status_list'],
                        STATUS_DELIMITERS
                        )
                    )

                ret['groups']['status'] = '.'.join(
                    ret['groups']['status_list']
                    )

                ret['groups']['status_dirty'] = \
                    ''.join(ret['groups']['status_list_dirty'])

            # extension

            ret['groups']['extension'] = extension

    if not isinstance(ret, dict):
        if not mute:
            logging.info("No match `{}'".format(filename))

        ret = None

    return ret


def parse_test():
    """
    Run parser on all difficult names (:data:`DIFFICULT_NAMES`) in test
    purposes
    """

    for i in DIFFICULT_NAMES:
        logging.info("====== Testing parser on `{}' ======".format(i))
        res = parse_tarball_name(i)
        if not isinstance(res, dict):
            logging.error(
                "Error parsing file name `{}' - parser not matched".format(i)
                )
        else:
            pprint.pprint(res)
            print()

    return


# TODO: adopt lists.filter_text_parse()
def filter_text_parse(filter_text):
    """
    Returns list of command structures

    ret = [
        dict(
            action   = '-' or '+',
            subject  = in ['path', 'filename', 'version', 'status'],
            function = <depends on subject> (no spaces allowed),
            data     = <depends on subject> (can contain spaces)
            )
        ]

    """
    ret = []

    lines = filter_text.splitlines()

    for i in lines:
        if i != '' and not i.isspace() and not i.startswith('#'):
            struct = i.split(' ', maxsplit=3)
            if not len(struct) == 4:
                logging.error("Wrong filter line: `{}'".format(i))
            else:
                struct = dict(
                    action=struct[0],
                    subject=struct[1],
                    function=struct[2],
                    data=struct[3]
                    )
                ret.append(struct)

    return ret


# TODO: adopt lists.filter_list()
def filter_tarball_list(input_list, filter_text):
    """
    Filter's supplied list with supplied filter text

    subjects not in check_for_subjects will always be positive (but can be
    filtered out by proper leading rules)
    """

    ret = []

    # TODO: do not use set function here as it is not sorted
    inp_list = set(copy.copy(input_list))
    out_list = copy.copy(inp_list)

    filters = filter_text_parse(filter_text)

    for f in filters:

        action = f['action']
        subject = f['subject']
        function = f['function']
        no = False
        data = f['data']

        if not action in ['+', '-']:
            logging.error("Wrong action: `{}'".format(action))
            ret = 10
            break

        if function.startswith('!'):
            no = True
            function = function[1:]

        if not subject in ['filename', 'version', 'status']:
            logging.error("Wrong subject : `{}'".format(subject))
            ret = 1
            break

        if subject in ['filename', 'status']:

            if not function in [
                    'begins', 'contains', 'ends', 'fm', 'bfm', 're']:
                logging.error(
                    "Wrong `{}' function : `{}'".format(subject, function)
                    )
                ret = 3
                break

        elif subject == 'version':

            if not function in [
                    '<', '<=', '==', '>=', '>', 're', 'fm',
                    'begins', 'contains', 'ends'
                    ]:
                logging.error(
                    "Wrong `version' function : `{}'".format(function)
                    )
                ret = 4
                break

        else:
            raise Exception("Programming error")

        if not isinstance(ret, int):

            working_list = None

            if action == '+':
                working_list = copy.copy(inp_list)

            elif action == '-':
                working_list = copy.copy(out_list)
            else:
                raise Exception("Programming Error")

            for item in working_list:

                working_item = item

                if subject == 'filename':
                    working_item = os.path.basename(item)

                elif subject in ['version', 'status']:

                    working_item = None

                    parsed = wayround_org.utils.tarball.\
                        parse_tarball_name(
                            os.path.basename(item),
                            mute=True
                            )

                    if not isinstance(parsed, dict):
                        # TODO: it's not error, but may be it's need to do
                        # something than just a `pass'
                        pass
                    else:
                        if subject == 'version':
                            working_item = parsed['groups']['version']

                        elif subject == 'status':
                            working_item = parsed['groups']['status']

                        else:
                            raise Exception("Programming error")

                else:
                    raise Exception("Programming error")

                matched = False

                if function == 'begins':
                    matched = working_item.startswith(data)

                elif function == 'contains':
                    matched = working_item.find(data) != -1

                elif function == 'end':
                    matched = working_item.endswith(data)

                elif function == 're':
                    matched = re.match(data, working_item) is not None

                elif function == 'fm':
                    logging.debug(
                        "filter_tarball_list: "
                        "fm-matching `{}' and `{}'".format(
                            working_item,
                            data
                            )
                        )
                    matched = fnmatch.fnmatch(working_item, data)

                elif function == 'bfm':
                    logging.debug(
                        "filter_tarball_list: "
                        "base fm-matching `{}' and `{}'".format(
                            working_item,
                            data
                            )
                        )
                    matched = fnmatch.fnmatch(
                        os.path.basename(working_item),
                        data
                        )

                elif function in ['<', '<=', '==', '>=', '>']:
                    matched = (
                        wayround_org.aipsetup.version.lb_comparator(
                            working_item,
                            function + ' ' + data
                            )
                        )
                else:
                    raise Exception("Programming error")

                if no:
                    matched = not matched

                if matched:

                    logging.debug(
                        "filter_tarball_list: "
                        "match: `{}'\n       `{}'".format(item, f)
                        )

                    if action == '+':
                        logging.debug(
                            "filter_tarball_list: adding: {}".format(item)
                            )
                        out_list.add(item)

                    elif action == '-':
                        logging.debug(
                            "filter_tarball_list: removing: {}".format(
                                item
                                )
                            )
                        if item in out_list:
                            out_list.remove(item)

                    else:
                        raise Exception("Programming error")

                else:
                    logging.debug(
                        "filter_tarball_list: NOT "
                        "match: `{}'\n       `{}'".format(item, f)
                        )

    if not isinstance(ret, int):
        ret = out_list

    if isinstance(ret, set):
        ret = list(ret)

    return ret





import subprocess
import os

import wayround_org.utils.file


def findtool(which=None, where=None):
    if which is None:

        which = 'pkg-config'

        if where is None:
            where = os.environ['PATH'].split(':')

        which = wayround_org.utils.file.which(
            'pkg-config',
            where
            )
    return which


def pkgconfig_include(names, which=None, where=None):
    ret = pkgconfig(names, '--cflags', which=which, where=where)
    s = ret.split()
    for i in range(len(s) - 1, -1, -1):
        if not s[i].startswith('-I'):
            del s[i]
    ret = ' '.join(s)
    return ret


def pkgconfig(names, options, which=None, where=None):

    which = findtool(which, where)

    if not isinstance(names, list):
        names = [names]

    if not isinstance(options, list):
        options = [options]

    cmd = [which] + options + names

    p = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
        )

    stdout, stderr = p.communicate()

    res = p.wait()

    ret = None

    if res != 0:
        pass

    else:
        stdout = str(stdout, 'utf-8')
        stdout = stdout.strip()

        ret = stdout

    return ret




import lxml.etree
import logging

import wayround_org.utils.types


def check_tagname_class_attrnames(tagname_class_attrnames):
    if not wayround_org.utils.types.struct_check(
            tagname_class_attrnames, {
                't': list, '.': {
                    't': tuple, '<': 3, '>': 6}}
            ):
        raise TypeError("`tagname_class_attrnames' has invalid structure")

    for i in range(len(tagname_class_attrnames)):

        for j in range(6 - len(tagname_class_attrnames[i])):
            new_val = tuple(tagname_class_attrnames[i]) + ('',)
            tagname_class_attrnames[i] = new_val

        if tagname_class_attrnames[i][3] is None:
            new_val = (i[0], i[1], i[2], '', i[4], i[5])
            tagname_class_attrnames[i] = new_val

        if tagname_class_attrnames[i][4] is None:
            new_val = (i[0], i[1], i[2], i[3], '', i[5])
            tagname_class_attrnames[i] = new_val

        if tagname_class_attrnames[i][5] is None:
            new_val = (i[0], i[1], i[2], i[3], i[4], '')
            tagname_class_attrnames[i] = new_val

        if not tagname_class_attrnames[i][3] in ['', '?', '+', '*']:
            raise ValueError(
                "invalid mode: `{}'".format(tagname_class_attrnames[i][3])
                )

    return


def subelems_to_object_props(
        element, obj, tagname_class_attrnames,
        must_be=False
        ):
    """
    Find element by tag and convert it to object using additional information

    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'), ...]

    return: False - no error, True - if must_be==True and some element not
        found
    """

    ret = False

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml tag element")

    check_tagname_class_attrnames(tagname_class_attrnames)

    for i in tagname_class_attrnames:

        val = element.find(i[0])
        if val is not None:
            getattr(obj, 'set_{}'.format(i[2]))(i[1].new_from_element(val))
        else:
            if must_be:
                ret = True

    return ret


def subelems_to_object_props2(element, obj, tagname_class_attrnames):
    """
    In distinction to subelems_to_object_props(), the forth values of tuples
    are not ignored and used to check the validity on element appearance.

    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'[, '[\?\*\+]?']), ...]
    """

    ret = False

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml tag element")

    check_tagname_class_attrnames(tagname_class_attrnames)

    must_be = []
    can_be = []
    multiples = []
    multiples_plus = []

    for i in tagname_class_attrnames:

        if i[3] == '?':
            must_be.append(i)

        elif i[3] == '*':
            multiples.append(i)

        elif i[3] == '+':
            multiples_plus.append(i)

        else:
            must_be.append(i)

    res = subelems_to_object_props(
        element,
        obj,
        tagname_class_attrnames=must_be,
        must_be=True
        )

    if res == True:
        ret = True

    else:
        subelems_to_object_props(
            element, obj, can_be
            )

        subelemsm_to_object_propsm(
            element, obj, multiples, must_be=False
            )

        subelemsm_to_object_propsm(
            element, obj, multiples_plus, must_be=True
            )

    return ret


def subelemsm_to_object_propsm(
        element, obj, tagname_class_attrnames, must_be=False
        ):
    """
    Find many elements by tag and convert them to objects using additional
    information

    must_be indicated necessarity of results to be len(lst) > 0

    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'[, '[\?\*\+]?']), ...]
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    check_tagname_class_attrnames(tagname_class_attrnames)

    for i in tagname_class_attrnames:

        gt_func = getattr(obj, 'get_{}'.format(i[2]))

        objs = gt_func()

        vals = element.findall(i[0])

        if must_be and len(vals) == 0:
            raise ValueError("Too few elements of `{}'".format(i[0]))

        for val in vals:
            objs.append(i[1].new_from_element(val))

        getattr(obj, 'check_{}'.format(i[2]))(objs)

    return


def elem_props_to_object_props(element, obj, names):
    """
    Get named element properties and apply them to object

    names must have following structure:
    [('element_prop_name', 'object_prop_name'), ...]
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    if not wayround_org.utils.types.struct_check(
            names,
            {'t': list, '.':
             {'t': tuple, '<': 2, '>': 2, '.': {'t': str}
              }
             }
            ):
        raise TypeError("`names' has invalid structure")

    for i in names:
        getattr(obj, 'set_{}'.format(i[1]))(element.get(i[0]))

    return


def object_props_to_subelems(obj, element, names):
    """
    Get named object properties and generate subelements to element

    names must be list of names of object properties
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    if not wayround_org.utils.types.struct_check(
            names,
            {'t': list, '.': {'t': str}}
            ):
        raise TypeError("`names' must be list of strings")

    for i in names:
        val = getattr(obj, 'get_{}'.format(i))()
        if val is not None:
            element.append(val.gen_element())

    return


def object_props_to_subelems2(obj, element, tagname_class_attrnames):
    """
    In distinction to object_props_to_subelems(), firsts values of tuples can
    end with '*', '+' or '?' , which indicates necessity to use function
    object_propsm_to_subelemsm()

    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'[, '[\?\*\+]?']), ...]
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    check_tagname_class_attrnames(tagname_class_attrnames)

    must_be = []
    multiples = []

    for i in tagname_class_attrnames:

        if i[3] == '?':
            must_be.append(i)

        elif i[3] == '*':
            multiples.append(i)

        elif i[3] == '+':
            multiples.append(i)

        else:
            must_be.append(i)

    object_props_to_subelems(
        obj,
        element,
        names=list(i[2] for i in must_be)
        )

    object_propsm_to_subelemsm(
        obj,
        element,
        names=list(i[2] for i in multiples)
        )

    return


def object_propsm_to_subelemsm(obj, element, names):
    """
    Get named object properties and generate subelements to element

    names must be list of names of object properties
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    if not wayround_org.utils.types.struct_check(
            names,
            {'t': list, '.': {'t': str}}
            ):
        raise TypeError("`names' must be list of strings")

    for i in names:
        vals = getattr(obj, 'get_{}'.format(i))()
        for val in vals:
            element.append(val.gen_element())

    return


def object_props_to_elem_props(obj, element, names):
    """
    Get named object properties and make element properties

    names must have following structure (seq of seqs of strings):
    [('object_prop_name', 'element_prop_name'), ...]
    """

    if not is_lxml_tag_element(element):
        raise TypeError("`element' must be lxml.etree.Element")

    if not wayround_org.utils.types.struct_check(
            names,
            {'t': list, '.':
             {'t': tuple, '<': 2, '>': 2, '.': {'t': str}}
             }
            ):
        raise TypeError("`names' has invalid structure")

    for i in names:

        val = getattr(obj, 'get_{}'.format(i[0]))()
        if val is not None:
            element.set(i[1], val)

    return


def subelems_to_order(element, order, tagname_class_attrnames):
    """
    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'[, '[\?\*\+]?']), ...]
    """

    check_tagname_class_attrnames(tagname_class_attrnames)

    for i in element:

        if is_lxml_tag_element(i):

            cls = None
            tag = None
            prop = None

            for j in tagname_class_attrnames:
                if (j[0] == i.tag):
                    cls = j[1]
                    tag = i.tag
                    prop = j[2]
                    break

            if cls is not None:
                try:
                    order.append((tag, cls.new_from_element(i), prop))
                except:
                    logging.error(
                        "Can't bring to order {}. ({}):\n--\n{}\n--\n".format(
                            i,
                            type(i),
                            lxml.etree.tostring(i)
                            )
                        )

    return


def order_to_subelems(order, element):

    for i in order:
        element.append(i[1].gen_element())

    return


def parse_tag(inp_str, localname, namespaces=None):

    if localname is None:
        localname = []

    if not isinstance(localname, list):
        localname = [localname]

    if isinstance(namespaces, str):
        namespaces = [namespaces]

    if not wayround_org.utils.types.struct_check(
            localname,
            {'t': list, '.': {'t': str}}
            ):
        raise TypeError("`localname' must be list of strings")

    if not wayround_org.utils.types.struct_check(
            namespaces,
            {'t': list, 'None': True, '.': {'t': str}}
            ):
        raise TypeError("`namespaces' has invalid structure")

    qname = lxml.etree.QName(inp_str)

    ret = None, None

    if len(localname) == 0 or qname.localname in localname:
        if namespaces is None:
            ret = qname.localname, qname.namespace
        else:
            if qname.namespace in namespaces:
                ret = qname.localname, qname.namespace

    return ret


def parse_element_tag(element, localname, namespaces=None):
    """
    Tag parse routine

    localname must be str or list of str

    returns  (localname, namespace) where namespace is one of namespaces or
    (None, None)
    """

    if not is_lxml_tag_element(element):
        raise ValueError(
            "`element' must be lxml tag element, not ({}): {}".format(
                type(element),
                element
                )
            )

    ret = parse_tag(element.tag, localname, namespaces)

    return ret


def checker_factory(cls, tagname_class_attrnames):
    """
    Factories check methods for pointed class using tagname_class_attrnames
    structure

    tagname_class_attrnames must have following structure:
    [('{ns}tag', ClassName, 'property_name'[, '[\?\*\+]?']), ...]
    """

    check_tagname_class_attrnames(tagname_class_attrnames)

    for i in tagname_class_attrnames:

        lst = False
        none = False
        if i[3] == '?':
            lst = False
            none = True
        elif i[3] == '+':
            lst = True
            none = False
        elif i[3] == '*':
            lst = True
            none = True
        else:
            lst = False
            none = False

        typ = i[1]

        check = ''
        doc = ''
        if lst == False and none == False:
            check = """
    if not type(value) == typ:
        raise TypeError(
            "`{}' must be of type {}, not {}".format(_name, typ, type(value))
            )
            """
            doc = 'value must be of type {}'.format(typ)

        elif lst == False and none == True:
            check = """
    if value != None and not type(value) == typ:
        raise TypeError(
    "`{}' must be None or of type {}, not {}".format(_name, typ, type(value))
            )
            """
            doc = 'value must be None or of type {}'.format(typ)

        elif lst == True and none == False:
            check = """
    if not wayround_org.utils.types.struct_check(
        value,
        {'t': list, '<': 1, '.': {'t': typ}}
        ):
        raise TypeError(
    "`{}' must be unempty list of type {}, not {}".format(_name, typ, type(value))
            )
"""
            doc = 'value must be unempty list of type {}'.format(typ)

        elif lst == True and none == True:
            check = """
    if not wayround_org.utils.types.struct_check(
        value,
        {'t': list, '.': {'t': typ}}
        ):
        raise TypeError(
    "`{}' must be list of type {}, not {}".format(_name, typ, type(value))
            )
"""
            doc = 'value must be list of type {}'.format(typ)
        else:
            raise Exception("programming error")

        exec("""
def check(self, value):
    import wayround_org.utils.types

    _name = '{name}'

{check}

cles.check_{name} = check
cles.check_{name}.__doc__ = doc

del check
        """.format(name=i[2], check=check),
             {'typ': typ, 'cles': cls, 'doc': doc,
              'tagname_class_attrnames': tagname_class_attrnames}
             )

    return


def simple_exchange_class_factory(
        cls,
        tag,
        namespace,
        subelements_struct,
        properties_list,
        value_name=None
        ):
    """
    Factories simple class structure to handle simple lxml elements

    see wayround_org.xmpp.xcard_4 for example usage

    Yes, this is not as beautiful as using class inheritance, but match simpler
    and obvious

    if value_name != None methods generated by this factory, requires following
    methods to be present in class at runtime:
    --
    set_{value_name}(value)
    get_{value_name}(value)
    --
    You can use
    wayround_org.utils.factory.class_generate_attributes_and_check()
    to generate set_this_element_text_value() and get_this_element_text_value()
    automatically
    """

    gw1 = ''
    gw2 = ''
    if value_name is not None:
        gw1 = 'cl.set_{value_name}(element.text)'.format(
            value_name=value_name
            )
        gw2 = 'el.text = cl.get_{value_name}()'.format(
            value_name=value_name
            )

    exec("""
import logging
import lxml.etree
import wayround_org.utils.lxml

def __init__(self, **kwargs):

    for i in self._subelements_struct:
        if i[3] in ['+', '*']:
            if not i[2] in kwargs or kwargs[i[2]] == None:
                kwargs[i[2]] = []

    for i in self._properties_list:
        set_func = getattr(self, 'set_{{}}'.format(i))
        set_func(kwargs.get(i))

    return

def new_from_element(cls, element):

    int_tag = wayround_org.utils.lxml.parse_element_tag(
        element,
        '{tag}',
        '{namespace}'
        )[0]

    if int_tag is None:
        raise ValueError("invalid element tag or namespace")

    cl = cls(**wayround_org.utils.lxml.\
        generate_initial_parameters_for_constructors(cls, element))

    {gw1}

    wayround_org.utils.lxml.subelems_to_object_props2(
        element, cl,
        cls._subelements_struct
        )

    return cl

def new_empty(cls, *args, **kwargs):
    cl = cls(**wayround_org.utils.lxml.\
        generate_empty_parameters_for_constructors(
            cls, *args, **kwargs)
            )
    return cl

def corresponding_tag(cls):
    return '{tag}'

corresponding_tag.__doc__ = \
    "returns tag name, for which this class generated"

def gen_element(self):

    self.check()

    el = lxml.etree.Element('{tag}')

    {gw2}

    wayround_org.utils.lxml.object_props_to_subelems2(
        self, el,
        self._subelements_struct
        )

    return el

clas._properties_list = PROPERTIES_LIST
clas._subelements_struct = SUBELEMENTS_STRUCT
clas.__init__ = __init__
clas.new_from_element = classmethod(new_from_element)
clas.new_empty = classmethod(new_empty)
clas.corresponding_tag = classmethod(corresponding_tag)
clas.gen_element = gen_element

del __init__
del new_from_element
del gen_element
del corresponding_tag

""".format(
        gw1=gw1,
        gw2=gw2,
        tag=tag,
        namespace=namespace
    ),
        {
        'PROPERTIES_LIST': properties_list,
        'SUBELEMENTS_STRUCT': subelements_struct,
        'clas': cls
    }
    )


# def is_lxml_tag_element(element):
#    return type(element) == lxml.etree._Element

def is_lxml_tag_element(element):
    return hasattr(element, 'tag') and isinstance(element.tag, str)


def generate_initial_parameters_for_constructors(cls, element):

    nec_params = {}

    for i in cls._subelements_struct:

        if i[3] == '':
            found = element.find(i[0])
            if found is not None:
                nec_params[i[2]] = i[1].new_from_element(found)
            else:
                raise ValueError(
                    "Not found required element `{{}}'".format(i[0])
                    )
        elif i[3] in ['*', '+']:
            nec_params[i[2]] = \
                [i[1].new_from_element(j) for j in element.findall(i[0])]
            if len(nec_params[i[2]]) == 0 and i[3] == '+':
                raise ValueError(
                    "Not found required element `{{}}'".format(i[0])
                    )

    return nec_params


def generate_empty_parameters_for_constructors(cls, *args, **kwargs):

    nec_params = {}

    for i in cls._subelements_struct:
        if i[3] == '':
            nec_params[i[2]] = i[1].new_empty(*args, **kwargs)
        elif i[3] in ['*', '+']:
            nec_params[i[2]] = []
            if len(nec_params[i[2]]) == 0 and i[3] == '+':
                nec_params[i[2]].append(i[1].new_empty(*args, **kwargs))

    return nec_params




"""
Version comparison utilities
"""

import logging
import os.path

import wayround_org.utils.tarball


def source_version_comparator(
        name1, name2,
        acceptable_source_name_extensions
        ):

    ret = 0

    if isinstance(acceptable_source_name_extensions, str):
        acceptable_source_name_extensions = \
            acceptable_source_name_extensions.split(' ')

    name1 = os.path.basename(name1)
    name2 = os.path.basename(name2)

    d1 = wayround_org.utils.tarball.parse_tarball_name(
        name1,
        True,
        acceptable_source_name_extensions
        )

    d2 = wayround_org.utils.tarball.parse_tarball_name(
        name2,
        True,
        acceptable_source_name_extensions
        )

    if d1 is None:
        raise Exception("Can't parse filename: {}".format(name1))

    if d2 is None:
        raise Exception("Can't parse filename: {}".format(name2))

    if d1['groups']['name'] != d2['groups']['name']:
        raise ValueError(
            "Files has different names: `{}' ({}) and `{}' ({})".format(
                d1['groups']['name'], name1,
                d2['groups']['name'], name2
                )
            )

    else:
        com_res = standard_comparison(
            d1['groups']['version_list'], d1['groups']['status_list'],
            d2['groups']['version_list'], d2['groups']['status_list']
            )

        if com_res != 0:
            ret = com_res
        else:
            ret = 0

    if ret == -1:
        logging.debug(name1 + ' < ' + name2)
    elif ret == 1:
        logging.debug(name1 + ' > ' + name2)
    else:
        logging.debug(name1 + ' = ' + name2)

    return ret


def standard_comparator(
        version1,
        version2
        ):

    logging.debug("standard_comparator: `{}', `{}'".format(version1, version2))

    int_v1 = version1
    int_v2 = version2

    i1_error = False
    i2_error = False

    if isinstance(version1, str):
        int_v1 = version1.split('.')

    if isinstance(version2, str):
        int_v2 = version2.split('.')

    if not isinstance(int_v1, list):
        i1_error = True
    else:
        for i in int_v1:
            if not isinstance(i, (int, str)):
                i1_error = True

    if not isinstance(int_v2, list):
        i2_error = True
    else:
        for i in int_v2:
            if not isinstance(i, (int, str)):
                i2_error = True

    if i1_error:
        raise ValueError(
            "standart_comparison parameters must be [lists of [str or int]]"
            " or [strings], not {}".format(int_v1)
            )

    if i2_error:
        raise ValueError(
            "standart_comparison parameters must be [lists of [str or int]]"
            " or [strings], not {}".format(int_v2)
            )

    ret = standard_comparison(int_v1, None, int_v2, None)
    logging.debug("standard_comparator ret: `{}'".format(ret))
    return ret


def standard_comparison(
        version_list1, status_list1,
        version_list2, status_list2
        ):

    vers_comp_res = None
    stat_comp_res = None

    vers1 = version_list1
    vers2 = version_list2

    longer = None

    v1l = len(vers1)
    v2l = len(vers2)

    #  length used in first comparison part
    el_1 = v1l

    if v1l == v2l:
        longer = None
        el_1 = v1l

    elif v1l > v2l:
        longer = 'vers1'
        el_1 = v2l

    else:
        longer = 'vers2'
        el_1 = v1l

    # first comparison part

    for i in range(el_1):

        if int(vers1[i]) > int(vers2[i]):
            logging.debug(vers1[i] + ' > ' + vers2[i])
            vers_comp_res = +1
            break
        elif int(vers1[i]) < int(vers2[i]):
            logging.debug(vers1[i] + ' < ' + vers2[i])
            vers_comp_res = -1
            break
        else:
            continue

    # second comparison part
    if vers_comp_res is None:
        if longer is not None:
            if longer == 'vers1':
                logging.debug(str(vers1) + ' > ' + str(vers2))
                vers_comp_res = +1
            else:
                logging.debug(str(vers1) + ' > ' + str(vers2))
                vers_comp_res = -1

    if vers_comp_res is None:
        vers_comp_res = 0

    if vers_comp_res == 0:
        if status_list1 is not None and status_list2 is not None:
            s1 = '.'.join(status_list1)
            s2 = '.'.join(status_list2)
            if s1 > s2:
                stat_comp_res = +1
            elif s1 < s2:
                stat_comp_res = -1
            else:
                stat_comp_res = 0

            vers_comp_res = stat_comp_res

    ret = vers_comp_res

    return ret




"""
License
=======

Python module for "walking" html
Copyright (C) 2011  Alexey V Gorshkov

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

About
=====
This is simple python module to parse html to A href values.

It can also do it recurcively accordingly to maxdepth parameter or as
deep as it can if maxdepth == -1

An html can be a file or a http server output.
"""

import xml.dom.minidom
import urllib
import re
import logging


def html_walk(html_text, maxdepth=10):
    """
    TODO
    """
    # TODO: todo
    pass


def list_rm_douplicates(lst=[]):
    """
    Routine for removal duplicated items from text list
    """
    lst2 = []

    for i in lst:
        if not i in lst2:
            lst2.append(i)

    return lst2


def html_link_get_links(url, tag_attr=[
        ('a', 'href'),
        # ('img', 'src'),
        # ('script', 'src')
        ]):
    """
    Tryes to open url, get it's content-type and if content-type is
    text/html - feed it to L{html_text_get_links}

    TODO

    @type url: text

    @param tag_attr: see L{dom_get_links}

    @return: link list on success or None on error

    @rtype: list or None
    """

    ret = None

    url_object = urllib.request.urlopen(url)
    ct = url_object.info().getheader('content-type').lower().strip()

    re_res = re.match(r'text/html(; *codepage=(.*))?', ct)

    if re_res is not None and re_res.group(
            1) is not None and re_res.group(2) is not None:

        text = url_object.read().decode(re_res.group(2))
        ret = html_text_get_links(text, tag_attr)

        if ret is None:
            logging.error(
                "Can't parse document " +
                url +
                " as XML (codepage:" +
                repr(
                    re_res.group(2)) +
                ")")

    # cleaningup for sure
    url_object.close()
    del url_object
    del re_res

    return ret


def html_text_get_links(html_text, base_url, tag_attr=[
        ('a', 'href'),
        # ('img', 'src'),
        # ('script', 'src')
        ]):
    """
    Parses text to xml.dom and feeds it to L{dom_get_links}

    Uses L{dom_get_links}, L{list_rm_douplicates} and
    L{list_rm_douplicates}.

    Returned list will be sorted alphabeticly.

    @type html_text: text

    @param base_url: used for expanding relative links

    @type base_url: text

    @param tag_attr: see L{dom_get_links}

    @return: list on success or None on error

    @rtype: list or None

    """
    docum = None
    try:
        docum = xml.dom.minidom.parseString(html_text)
    except:
        # print "-e- Can't parse text as XML"
        return None

    root = docum.documentElement
    lst = sorted(dom_get_links(root))
    lst = list_rm_douplicates(lst)
    return lst

# def expand_links(lst=[], cwd=''):
#     """
#     Not completed
#     """
#     lst2 = []

#     for i in lst:
#         if not i in lst2:
#             lst2.append(i)

#     return lst2


def dom_get_links(root, tag_attr=[
        ('a', 'href'),
        # ('img', 'src'),
        # ('script', 'src')
        ]):
    """
    Get all <a href=""> links from domain xml.dom objects recurcively

    @param root: xml.dom object

    @param tag_attr: list of tuples, each of which has two values:

       0. tagname for which to look

       1. this tag attribute name which to add to returnable list.

    @return: list with all links in xml.dom object

    @rtype: list
    """
    lst = []

    for i in root.childNodes:
        if i.nodeType == xml.dom.Node.ELEMENT_NODE:

            if len(i.childNodes) > 0:
                lst += dom_get_links(i)

            for tag, attr in tag_attr:

                if i.tagName == tag:
                    if i.hasAttribute(attr):
                        t = i.getAttribute(attr)
                        lst.append(t)
                        del t

    return lst




import urllib.parse

import regex

import wayround_org.utils.types

URL_AUTHORITY_RE = r'^((?P<userinfo>.*)\@)?(?P<host>.*)(\:(?P<port>\d+))?$'
URL_AUTHORITY_RE_C = regex.compile(URL_AUTHORITY_RE)

URI_RE = (
    r'^('
    r'(?P<scheme>\p{L}\w*\:)'
    r'(?P<path>.*?)'
    r'(\?(?P<query>.*))?'
    r'(?P<fragment>\#.*)?'
    r')$'
    )

URI_RE_C = regex.compile(URI_RE)


class UserInfoLikeHttp:

    def __init__(self, value):

        if isinstance(value, AuthorityLikeHttp):
            value = value.userinfo

        if not isinstance(value, str):
            raise TypeError("`value' must be str")

        self.name = value
        self.password = None

        if ':' in self.name:
            self.name, self.password = self.name.split(':', 1)

        return

    def __str__(self):

        ret = self.name
        if self.password is not None:
            ret = '{}:{}'.format(ret, self.password)

        return ret


class AuthorityLikeHttp:

    @classmethod
    def new_from_string(cls, value):

        res = URL_AUTHORITY_RE_C.match(value)

        if res is None:
            raise ValueError("can't parse value as authority")

        userinfo = res.group('userinfo')
        host = res.group('host')
        port = res.group('port')

        return cls(userinfo, host, port)

    def __init__(self, userinfo, host, port):

        self._userinfo = None
        self._host = None
        self._port = None

        self.userinfo = userinfo
        self.host = host
        self.port = port
        return

    def render_str(self):

        ret = ''

        if self.userinfo is not None:
            ret += '{}@'.format(self.userinfo)

        ret += self.host

        if self.port is not None:
            ret += ':{}'.format(self.port)

        return ret

    # don't do this. code should be more strict
    # render_string = render_str

    def __str__(self):
        return self.render_str()

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, value):
        if not isinstance(value, str):
            raise TypeError("`host' must be str")
        self._host = value.lower()
        return

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, value):
        if isinstance(value, str):
            value = int(value)

        if value is not None and not isinstance(value, int):
            raise TypeError("`port' must be int")

        self._port = value
        return

    @property
    def userinfo(self):
        return self._userinfo

    @userinfo.setter
    def userinfo(self, value):

        if isinstance(value, str):
            value = UserInfoLikeHttp(value)

        if value is not None and not isinstance(value, UserInfoLikeHttp):
            raise TypeError("`userinfo' must be None or UserInfoLikeHttp")

        self._userinfo = value
        return


class QueryLikeHttp:

    def __init__(self, value, encoding=None, errors=None):
        self._value = None
        self.value = value

        if encoding is None:
            encoding = 'utf-8'

        self._encoding = encoding
        self._errors = errors
        return

    def render_str(self):

        if encoding is None:
            encoding = 'utf-8'

        ret = []

        for i in self.value:
            ret.append(
                '{}={}'.format(
                    urllib.parse.quote(
                        i[0],
                        encoding=self._encoding,
                        errors=self._errors
                        ),
                    urllib.parse.quote(
                        i[1],
                        encoding=self._encoding,
                        errors=self._errors
                        )
                    )
                )

        ret = '&'.join(ret)

        return ret

    def __str__(self):
        return self.render_str()

    def __getitem__(self, index):
        return self.value[index]

    def __len__(self):
        return len(self.value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):

        if isinstance(value, URI):
            value = value.query

        if isinstance(value, str):
            res_val = []
            for i in value.split('&'):
                k, v = i.split('=', 1)

                k = urllib.parse.unquote(
                    k,
                    encoding=self._encoding,
                    errors=self._errors
                    ),
                v = urllib.parse.unquote(
                    v,
                    encoding=self._encoding,
                    errors=self._errors
                    )

                res_val.append(tuple())

            value = res_val
            del res_val

        if not wayround_org.utils.types.struct_check(
                value,
                {'t': list,
                 'None': True,
                 '.': {
                     't': tuple,
                     '<': 2,
                     '>': 2,
                     '.': {
                         't': str
                         }
                     }
                 }
                ):
            raise TypeError("`query' must be list of 2-str-tuples")

        self._value = value
        return

    def keys(self):
        ret = []
        for i in self.value:
            ret.append(i[0])
        return ret

    def values(self):
        ret = []
        for i in self.value:
            ret.append(i[1])
        return ret

    def get(self, key, default=None):
        ret = default
        for i in self.value:
            if i[0] == key:
                ret = i[1]
        return ret

    def get_all(self, key):
        ret = []
        for i in self.value:
            if i[0] == key:
                ret.append(i[1])
        return ret

    def set(self, key, value):
        val = self.value
        for i in range(len(val) - 1, -1, -1):
            if val[i][0] == key:
                val[i] = tuple(key, value)
        return

    def remove(self, key):
        val = self.value
        for i in range(len(val) - 1, -1, -1):
            if val[i][0] == key:
                del val[i]
        return


class URI:

    @classmethod
    def new_from_string(cls, value):

        res = URI_RE_C.match(value)

        if res is None:
            raise ValueError("can't parse value as URI string")

        scheme = res.group('scheme')
        authority = None
        # authority = res.group('authority')
        path = res.group('path')
        query = res.group('query')
        fragment = res.group('fragment')

        if path.startswith('//'):
            authority = None
            path_splitted = path[2:].split('/')
            authority = path_splitted[0]
            path = '/'.join(path_splitted[1:])

        return cls(scheme, authority, path, query, fragment)

    def __init__(self, scheme, authority, path, query, fragment):

        self._scheme = None
        self._authority = None
        self._path = None
        self._query = None
        self._fragment = None

        self.scheme = scheme
        self.authority = authority
        self.path = path
        self.query = query
        self.fragment = fragment

        return

    def render_str(self):
        ret = ''
        ret += str(self.scheme)
        ret += ':'
        if self.authority is not None:
            ret += str(self.authority)
        if self.path is not None:
            ret += str(self.path)
        if self.fragment is not None:
            ret += '#'
            ret += str(self.fragment)
        return ret

    def __str__(self):
        return self.render_str()

    def __repr__(self):
        ret = '{} {}'.format(
            repr(super()),
            ("scheme: `{}', authority: `{}', "
             "path: `{}', query: `{}', fragment: `{}'").format(
                self.scheme,
                self.authority,
                self.path,
                self.query,
                self.fragment
                )
            )
        return ret

    @property
    def scheme(self):
        return self._scheme

    @scheme.setter
    def scheme(self, value):
        if not isinstance(value, str):
            raise ValueError("`scheme' must be str")
        self._scheme = value.rstrip(':')
        # NOTE: scheme is case sensitive
        # self._scheme = self._scheme.lower()
        return

    @property
    def authority(self):
        return self._authority

    def gen_authority_like_http(self):
        return AuthorityLikeHttp(self.authority)

    @authority.setter
    def authority(self, value):
        if value is not None:
            a_t = type(value)

            if a_t == str:
                value = AuthorityLikeHttp.new_from_string(value)

            elif a_t == AuthorityLikeHttp:
                pass

            else:
                raise TypeError("invalid `authority' type")

        self._authority = value

        return

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):

        if value is None:
            value = []

        if isinstance(value, str):

            if self.scheme.lower() == 'urn':
                value = value.split(':')
            else:
                value = value.split('/')

        if not wayround_org.utils.types.struct_check(
                value,
                {'t': list,
                 'None': True,
                 '.': {
                     't': str
                     }
                 }
                ):
            raise TypeError("`path' must be list of strings")

        self._path = value

        return

    @property
    def query(self):
        return self._query

    def gen_query_like_http(self):
        return QueryLikeHttp(self.query)

    @query.setter
    def query(self, value):

        if not value is None and not isinstance(value, str):
            raise TypeError("`query' must be None or str")

        self._query = value

        return

    @property
    def fragment(self):
        return self._fragment

    @fragment.setter
    def fragment(self, value):

        if value is not None and not isinstance(value, str):
            raise ValueError("`fragment' must be str")

        self._fragment = value

        return


class HttpURI(URI):
    """
    This class is specially done for HTTP URIs handling, as it is very
    common case.

    .query and .authority are stored as objects, not as strings
    """

    def gen_authority_like_http(self):
        raise Exception("not in HttpURI")

    def gen_query_like_http(self):
        raise Exception("not in HttpURI")

    @property
    def authority(self):
        return super().authority

    @authority.setter
    def authority(self, value):

        if isinstance(value, str):
            value = AuthorityLikeHttp2.new_from_string(value)

        if value is not None and not isinstance(value, AuthorityLikeHttp):
            raise ValueError(
                "`query' must be None or AuthorityLikeHttpp6 inst"
                )

        self._authority = value

        return

    @property
    def query(self):
        return super().query

    @query.setter
    def query(self, value):

        if isinstance(value, str):
            value = QueryLikeHttp(value)

        if value is not None and not isinstance(value, QueryLikeHttp):
            raise ValueError("`query' must be None or QueryLikeHttp inst")

        self._query = value

        return

    def is_email(self):

        has_scheme = isinstance(self.scheme, str)

        has_user_info = False
        has_password = False
        has_domain = False
        has_port = False

        if self.authority is not None:
            a = self.authority
            has_domain = isinstance(a.host, str)
            has_port = isinstance(a.port, int)
            if a.userinfo is not None:
                a = a.userinfo
                has_user_info = True
                has_password = a.password is not None

        has_query = self.query is not None
        has_fragment = self.fragment is not None

        ret = (
            not has_scheme
            and has_user_info
            and not has_password
            and has_domain
            and not has_port
            and not has_query
            and not has_fragment
            )
        return ret


def run_examples():

    examples = [
        'ftp://ftp.is.co.za/rfc/rfc1808.txt',
        'http://www.ietf.org/rfc/rfc2396.txt',
        'ldap://[2001:db8::7]/c=GB?objectClass?one',
        'mailto:John.Doe@example.com',
        'news:comp.infosystems.www.servers.unix',
        'tel:+1-816-555-1212',
        'telnet://192.0.2.16:80/',
        'urn:oasis:names:specification:docbook:dtd:xml:4.1.2'
        ]

    for i in examples:
        print("{}\n\t{}\n".format(i, repr(URI.new_from_string(i))))




"""
Look for easy usage example in tag.py moule

thanks to python language flexibility, I can create such modules and classes,
without creating singletones in modules
"""

import sqlalchemy.ext.declarative
import sqlalchemy.orm


class BasicDB:

    def __init__(
            self,
            config_string=None,
            bind=None,
            decl_base=None,
            metadata=None,
            init_table_data=None
            ):

        if bind is None and config_string is None:
            raise ValueError(
                "if `bind' or `config_string' must be defined"
                )

        if bind is not None and config_string is not None:
            raise ValueError(
                "if `bind' is defined, then `config_string' must be None"
                )

        if config_string is not None and bind is None:
            bind = sqlalchemy.create_engine(config_string)

        if decl_base is None:
            self.decl_base = sqlalchemy.ext.declarative.declarative_base(
                metadata=metadata,
                bind=bind
                )
        else:
            self.decl_base = decl_base

        self.init_table_mappings(init_table_data)

        return

    def create_all(self):
        return self.decl_base.metadata.create_all()

    def bind(self, engine):
        self.decl_base.metadata.bind = engine
        return




import io
import logging
import os.path
import shutil
import subprocess
import sys
import tarfile
import lzma

import wayround_org.utils.exec
import wayround_org.utils.file
import wayround_org.utils.path
import wayround_org.utils.stream


CANONICAL_COMPRESSORS = frozenset(['xz', 'lzma', 'bzip2', 'gzip'])


def _extract_zip(file_name, output_dir):

    ret = 0

    try:
        proc = wayround_org.utils.exec.simple_exec(
            'unzip', stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=sys.stderr,
            options=['-qq', '-o', file_name, '-d', output_dir],
            )
    except:
        logging.exception("unzip start error")
        ret = 1
    else:
        ret = proc.wait()

    return ret


def _extract_7z(file_name, output_dir):

    ret = 0

    try:
        proc = wayround_org.utils.exec.simple_exec(
            '7z', stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=sys.stderr,
            options=['x', '-y', '-o{}'.format(output_dir), file_name],
            )
    except:
        logging.exception("unzip start error")
        ret = 1
    else:
        ret = proc.wait()

    return ret


def _extract_tar_7z(file_name, output_dir):

    ret = 0

    try:
        proc_7z = wayround_org.utils.exec.simple_exec(
            '7z',
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=None,
            options=['x', '-so', file_name],
            )
    except:
        logging.exception("7z start error")
        ret = 1

    else:

        try:
            proc_tar = wayround_org.utils.exec.simple_exec(
                'tar',
                stdin=proc_7z.stdout,
                stdout=subprocess.PIPE,
                stderr=None,
                options=[
                    '--no-same-owner',
                    '--no-same-permissions',
                    '-xlRC',
                    output_dir
                    ]
                )
            proc_7z.wait()
        except:
            logging.exception("tar start error")
            ret = 2
        else:
            proc_tar.wait()

        finally:
            if proc_tar.returncode is None:
                proc_tar.terminate()

    return ret


def _extract_tar_arch(file_name, output_dir, compressor, log=None):

    # TODO: cleanups and reworks required

    os.makedirs(output_dir, exist_ok=True)

    out = None
    err = None
    if log is not None:
        out = log.stdout
        err = log.stderr

    p = subprocess.Popen(
        ['tar', '-xf', file_name, '-C', output_dir, '-v'],
        stderr=err,
        stdout=out
        )
    ret = p.wait()
    return ret

    '''
    if not compressor in CANONICAL_COMPRESSORS:
        raise ValueError(
            "compressor not in `{}'".format(CANONICAL_COMPRESSORS)
            )

    ret = extract_tar_canonical(
        file_name,
        output_dir,
        compressor,
        verbose_tar=True,
        verbose_compressor=True,
        log=log,
        add_tar_options=[
            '--no-same-owner',
            '--no-same-permissions'
            ]
        )

    return ret
    '''


def extract(file_name, output_dir, log=None):

    ret = None

    if file_name.endswith('.tar.lzma'):
        ret = _extract_tar_arch(file_name, output_dir, 'lzma', log=log)

    elif file_name.endswith('.tar.bz2'):
        ret = _extract_tar_arch(file_name, output_dir, 'bzip2', log=log)

    elif file_name.endswith('.tar.gz'):
        ret = _extract_tar_arch(file_name, output_dir, 'gzip', log=log)

    elif file_name.endswith('.tar.xz'):
        ret = _extract_tar_arch(file_name, output_dir, 'xz', log=log)

    elif file_name.endswith('.tar.7z'):
        ret = _extract_tar_7z(file_name, output_dir)

    elif file_name.endswith('.tbz2'):
        ret = _extract_tar_arch(file_name, output_dir, 'bzip2', log=log)

    elif file_name.endswith('.tgz'):
        ret = _extract_tar_arch(file_name, output_dir, 'gzip', log=log)

    elif file_name.endswith('.zip'):
        ret = _extract_zip(file_name, output_dir)

    elif file_name.endswith('.7z'):
        ret = _extract_7z(file_name, output_dir)

    else:
        logging.error("Unsupported extension")

    if ret is None:
        raise Exception("Not implemented")

    return ret


def determine_compressor_by_filename(file_name, mute=False):

    ret = None

    if file_name.endswith('.lzma'):
        ret = 'xz'

    elif file_name.endswith('.bz2'):
        ret = 'bzip2'

    elif file_name.endswith('.gz'):
        ret = 'gzip'

    elif file_name.endswith('.xz'):
        ret = 'xz'

    else:
        if not mute:
            logging.error("Unknown compressor {}".format(file_name))

    return ret


def determine_extension_by_filename(file_name, mute=False):

    ret = None

    for i in ['.lzma', '.bz2', '.gz', '.xz']:

        if file_name.endswith(i):
            ret = i
            break

    if not ret:
        if not mute:
            logging.error("Unknown compressor extension {}".format(file_name))

    return ret


def canonical_compressor(
        compressor,
        stdin=None,
        stdout=None,
        stderr=None,
        options=None,
        ):
    """
    Canonical compressor shortcut

    Works only with `{}' compressors which
    must support -0 .. -9 , -v and -d options in canonical way
    """.format(repr(list(CANONICAL_COMPRESSORS)))

    # print("canonical_compressor: {}".format())

    if options is None:
        options = []

    if not compressor in CANONICAL_COMPRESSORS:
        raise ValueError(
            "canonical_compressor supports only"
            " `{}', but `{}' requested".format(
                list(CANONICAL_COMPRESSORS),
                compressor
                )
            )

    '''
    c = subprocess.Popen(
        [compressor] + options,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr
        )

    ret = c.wait()
    '''
    # while True:

    #'''
    ret = wayround_org.utils.exec.process_stream(
        compressor,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        options=options
        )
    #'''

    return ret


def archive_tar_canonical(
        dirname, output_filename,
        compressor,
        verbose_tar=False,
        verbose_compressor=False,
        bufsize=(2 * 1024 ** 2),
        additional_tar_options=None
        ):

    ret = 0
    try:
        fobj = open(output_filename, 'wb')
    except:
        logging.exception("Error opening file for write")
        ret = 1
    else:
        try:
            ret = archive_tar_canonical_fobj(
                dirname,
                fobj,
                compressor,
                verbose_tar,
                verbose_compressor,
                bufsize=bufsize,
                additional_tar_options=additional_tar_options
                )
        except:
            logging.exception("Error")
            ret = 2
    finally:
        fobj.close()

    return ret


def archive_tar_canonical_fobj(
        dirname,
        output_fobj,
        compressor,
        verbose_tar=False,
        verbose_compressor=False,
        bufsize=(2 * 1024 ** 2),
        additional_tar_options=None
        ):

    if additional_tar_options is None:
        additional_tar_options = []

    ret = 0

    dirname = wayround_org.utils.path.abspath(dirname)

    if not os.path.isdir(dirname):
        logging.error("Not a directory: {}".format(dirname))
        ret = 1
    else:

        options = []
        stderr = None

        if verbose_tar:
            options += ['-v']
            stderr = sys.stderr

        options += ['-c', '.']

        options += additional_tar_options

        try:
            tarproc = wayround_org.utils.exec.simple_exec(
                'tar',
                options=options,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                cwd=dirname,
                bufsize=bufsize,
                stderr=stderr
                )
        except:
            logging.exception("tar start error")
            ret = 2
        else:

            try:
                options = []
                stderr = None

                if verbose_compressor:
                    options += ['-v']
                    stderr = sys.stderr

                options += ['-9']

                if canonical_compressor(
                        compressor,
                        options=options,
                        stdin=tarproc.stdout,
                        stdout=output_fobj,
                        stderr=stderr
                        ) != 0:
                    ret = 3
                tarproc.wait()
            except:
                logging.exception("Error")

        finally:
            tarproc.poll()
            if tarproc.returncode is None:
                tarproc.terminate()

    return ret


def extract_tar_canonical(
        input_filename,
        dirname,
        compressor,
        verbose_tar=False,
        verbose_compressor=False,
        add_tar_options=None,
        log=None
        ):

    if not compressor in CANONICAL_COMPRESSORS:
        raise ValueError(
            "compressor not in `{}'".format(CANONICAL_COMPRESSORS)
            )

    if add_tar_options is None:
        add_tar_options = []

    ret = 0
    try:
        fobj = open(input_filename, 'rb')
    except:
        logging.exception("Error opening file for read")
        ret = 1
    else:
        try:
            ret = extract_tar_canonical_fobj(
                fobj,
                dirname,
                compressor,
                verbose_tar,
                verbose_compressor,
                add_tar_options,
                log=log
                )
        finally:
            fobj.close()
    return ret


def extract_tar_canonical_fobj(
        input_fobj,
        dirname,
        compressor,
        verbose_tar=False,
        verbose_compressor=False,
        add_tar_options=[],
        log=None
        ):

    if not compressor in CANONICAL_COMPRESSORS:
        raise ValueError(
            "compressor not in `{}'".format(CANONICAL_COMPRESSORS)
            )

    if add_tar_options is None:
        add_tar_options = []

    dirname = wayround_org.utils.path.abspath(dirname)

    ret = wayround_org.utils.file.create_if_not_exists_dir(dirname)

    if ret != 0:
        logging.error(
            "Error while checking destination dir: {}".format(dirname)
            )
        ret = 1
    else:

        options = [] + add_tar_options

        if verbose_tar:
            options += ['-v']

        options += ['-x']

        tarproc = None
        try:
            sout = sys.stdout

            if log is not None:
                # if False:
                sout = log.stdout
            tarproc = wayround_org.utils.exec.simple_exec(
                "tar",
                options=options,
                stdin=subprocess.PIPE,
                stdout=sout,
                cwd=dirname,
                stderr=sout
                )
        except:
            logging.exception("tar error detected")
            ret = 2
        else:

            try:
                options = []

                if verbose_compressor:
                    options += ['-v']

                options += ['-d']

                if canonical_compressor(
                        compressor,
                        stdin=input_fobj,
                        stdout=tarproc.stdin,
                        options=options,
                        stderr=sys.stderr
                        ) != 0:

                    ret = 3

                # if log is not None:
                # if False:
                #    log.input_from_Popen_process(tarproc)

                ret = tarproc.wait()

            finally:
                if tarproc.returncode is None:
                    tarproc.terminate()

    return ret


def pack_dir_contents_tar(
        dirname,
        output_filename,
        verbose_tar=False
        ):

    ret = 0

    dirname = wayround_org.utils.path.abspath(dirname)

    if not os.path.isdir(dirname):
        logging.error("Not a directory: {}".format(dirname))
        ret = 1
    else:
        try:
            outf = open(output_filename, 'wb')
        except:
            logging.exception(
                "Coundn't open `{}' for write".format(output_filename)
                )
            ret = 2
        else:

            try:
                options = []
                stderr = subprocess.PIPE

                if verbose_tar:
                    options += ['-v']
                    stderr = sys.stderr

                options += ['-c', '.']

                tarproc = wayround_org.utils.exec.simple_exec(
                    "tar",
                    options=options,
                    stdin=None,
                    stdout=outf,
                    cwd=dirname,
                    bufsize=2 * 1024 ** 2,
                    stderr=stderr
                    )

                tarproc.wait()
            finally:
                outf.close()
    return ret


def tar_get_member(tarobject, cont_name):
    ret = None

    try:
        ret = tarobject.getmember(cont_name)
    except KeyError:
        logging.error("Can't get tar member: {}".format(cont_name))
        ret = 2
    except:
        logging.exception("Can't get tar member: {}".format(cont_name))
        ret = 1

    return ret


def tar_member_extract_file(tarobject, member):
    ret = None

    try:
        ret = tarobject.extractfile(member)
    except:
        logging.exception("Can't get tar member")
        ret = 1

    return ret


def tar_member_get_extract_file(tarobject, cont_name):

    ret = None

    member = tar_get_member(tarobject, cont_name)

    if not isinstance(member, tarfile.TarInfo):
        ret = 1
    else:
        fileobj = tar_member_extract_file(tarobject, member)

        if not isinstance(fileobj, tarfile.ExFileObject):
            ret = 2
        else:
            ret = fileobj

    return ret


def tar_member_get_extract_file_to(tarf, cont_name, output_filename):

    ret = 0
    try:
        fd = open(output_filename, 'wb')
    except:
        logging.error("Error creating output file {}".format(output_filename))
        ret = 1
    else:
        try:
            fobj = tar_member_get_extract_file(
                tarf, cont_name
                )
            try:
                if not isinstance(fobj, tarfile.ExFileObject):
                    logging.error(
                        "Error getting {} from tar".format(cont_name)
                        )
                    ret = 2
                else:
                    wayround_org.utils.stream.cat(fobj, fd)
            finally:
                fobj.close()
        finally:
            fd.close()

    return ret


def xzcat(stdin, convert_to_str=None):

    if convert_to_str == True:
        convert_to_str = 'utf-8'
    elif convert_to_str == False:
        convert_to_str = None

    ret = 0

    dec = lzma.LZMADecompressor(memlimit=200 * 1024 ** 2)

    outstr = None
    if not convert_to_str:
        outstr = io.BytesIO()
    else:
        outstr = io.StringIO()

    dec_data = b''

    while True:

        if dec.eof:
            break

        try:
            buff = stdin.read(2 * 1024 ** 2)

            if len(buff) != 0:

                dec_data = dec.decompress(buff)

                if not convert_to_str:
                    outstr.write(dec_data)
                else:
                    outstr.write(str(dec_data, convert_to_str))

        except:
            logging.exception("Exception while decompressing LZMA data")
            ret = 1
            break

    if ret == 0:
        ret = outstr.getvalue()

    outstr.close()

    return ret


def extract_low(
        log,
        tmpdir,
        tarball,
        outdir,
        unwrap_dir=False,
        rename_dir=False,
        more_when_one_extracted_ok=False
        ):

    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    if not os.path.isdir(tmpdir):
        os.makedirs(tmpdir)

    if log:
        log.info("Extracting {}".format(os.path.basename(tarball)))

    extr_error = extract(
        tarball, tmpdir, log
        )

    ret = extr_error

    if extr_error != 0:
        if log:
            log.error(
                "Extraction error: {}".format(extr_error)
                )
        ret = 3
    else:

        extracted_dir = os.listdir(tmpdir)

        if len(extracted_dir) > 1 and not more_when_one_extracted_ok:
            if log:
                log.error("too many extracted files")
            ret = 4
        else:

            for i in extracted_dir:

                i2 = tmpdir + os.path.sep + i

                if unwrap_dir:

                    i2_files = os.listdir(i2)
                    for i in i2_files:
                        shutil.move(i2 + os.path.sep + i, outdir)
                    shutil.rmtree(i2)

                else:
                    if rename_dir:
                        n = outdir + os.path.sep + str(rename_dir)
                        if log:
                            log.info(
                                "moving extracted {}\n    as `{}'".format(
                                    i2,
                                    n
                                    )
                                )
                        shutil.move(i2, n)
                    else:
                        if log:
                            log.info(
                                "moving extracted {}\n    to `{}'".format(
                                    i2,
                                    outdir
                                    )
                                )

                        if os.path.isfile(i2):
                            new_file_name = os.path.join(
                                outdir,
                                os.path.basename(i2)
                                )

                            if os.path.isfile(new_file_name):
                                os.unlink(new_file_name)

                        shutil.move(i2, outdir)

    return ret


def tarobj_check_member_sum(tarobj, sums, member_name):
    """
    Check tarball member checksum.

    Sums must be supplied with sums parameter, which must be dict of building::

        sums == {'filename':'checksum'}
    """

    ret = True

    fobj = tar_member_get_extract_file(
        tarobj,
        member_name
        )

    if not isinstance(fobj, tarfile.ExFileObject):
        ret = False
    else:

        summ = wayround_org.utils.checksum.make_fileobj_checksum(fobj)

        if summ == sums[member_name]:
            ret = True
        else:
            ret = False

        fobj.close()

    return ret




import datetime
import wayround_org.utils.datetime_iso8601

TIMESTAMP_STR_FORMAT_PATTERN = \
    '{year:04d}{month:02d}{day:02d}.' \
    '{hour:02d}{minute:02d}{second:02d}.' \
    '{micro:07d}'

TIMESTAMP_RE_PATTERN = \
    r'(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})' \
    r'\.(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})\.(?P<micro>\d{7})'


TIMESTAMP_NO_POINTS_STR_FORMAT_PATTERN = \
    '{year:04d}{month:02d}{day:02d}' \
    '{hour:02d}{minute:02d}{second:02d}' \
    '{micro:07d}'

TIMESTAMP_NO_POINTS_RE_PATTERN = \
    r'(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})' \
    r'(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})(?P<micro>\d{7})'


def currenttime_stamp():
    d = datetime.datetime.now()
    return time_stamp(d)


def currenttime_stamp2():
    d = datetime.datetime.now()
    return time_stamp_no_points(d)


def currenttime_stamp_utc():
    d = datetime.datetime.utcnow()
    return time_stamp(d)


def currenttime_stamp2_utc():
    d = datetime.datetime.utcnow()
    return time_stamp_no_points(d)


current_timestamp = currenttime_stamp


def time_stamp(dt):
    return TIMESTAMP_STR_FORMAT_PATTERN.format_map(
        {
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
            'hour': dt.hour,
            'minute': dt.minute,
            'second': dt.second,
            'micro': dt.microsecond
            }
        )


def time_stamp_no_points(dt):
    return TIMESTAMP_NO_POINTS_STR_FORMAT_PATTERN.format_map(
        {
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
            'hour': dt.hour,
            'minute': dt.minute,
            'second': dt.second,
            'micro': dt.microsecond
            }
        )


def currenttime_stamp_iso8601():
    d = datetime.datetime.now()
    return time_stamp_iso8601(d)


def currenttime_stamp_iso8601_utc():
    d = datetime.datetime.utcnow()
    return time_stamp_iso8601(d)


def time_stamp_iso8601(dt):
    return wayround_org.utils.datetime_iso8601.datetime_to_str(dt)





def get_column_names(metadata, tablename):

    length = len(tablename) + 1

    ret = []

    for i in metadata.tables[tablename].columns:
        # FIXME: need to find more correct way than str()
        #        ret.append(str(i)[length:])
        ret.append(i.name)

    return ret




import copy
import os.path

import wayround_org.utils.types

S_SEP = os.path.sep
D_SEP = S_SEP * 2

S_SEP_BYTES = bytes(S_SEP, 'ascii')
D_SEP_BYTES = bytes(D_SEP, 'ascii')

# TODO: documentation for all functions
# TODO: make all functions support bytes as well as str


def select_seps(str_in):
    if not isinstance(str_in, (str, bytes)):
        raise ValueError("str_in must be str or bytes")

    ret = S_SEP, D_SEP
    if isinstance(str_in, bytes):
        ret = S_SEP_BYTES, D_SEP_BYTES
    return ret


def select_s_sep(str_in):
    ret = S_SEP
    if isinstance(str_in, bytes):
        ret = S_SEP_BYTES
    return ret


def _remove_double_sep(str_in):

    if not isinstance(str_in, (str, bytes)):
        raise ValueError("str_in must be str or bytes")

    if isinstance(str_in, str):
        while D_SEP in str_in:
            str_in = str_in.replace(D_SEP, S_SEP)

    elif isinstance(str_in, bytes):
        while D_SEP_BYTES in str_in:
            str_in = str_in.replace(D_SEP_BYTES, S_SEP_BYTES)

    return str_in


def _remove_trailing_slash(str_in):

    if not isinstance(str_in, (str, bytes)):
        raise ValueError("str_in must be str or bytes")

    s_sep = select_s_sep(str_in)

    ret = str_in

    while ret.endswith(s_sep):
        ret = ret[:-1]

    # TODO: what is it? commented out 2014-08-30
    # if len(ret) == 0:
    #     ret = s_sep

    return ret


def join(*args, must_be=str):

    if not must_be in [str, bytes]:
        raise ValueError("`must_be' must be in [str, bytes]")

    s_sep = S_SEP
    if must_be == bytes:
        s_sep = S_SEP_BYTES

    for i in args:
        if not isinstance(i, (str, bytes, list,)):
            raise ValueError("arguments must be str, bytes or lists")

    res, res_t = wayround_org.utils.types.is_all_rec_bytes_str(
        args,
        must_be=must_be
        )
    if not res:
        raise ValueError(
            "arguments must all be of same type: str, bytes or lists of them"
            )

    if len(args) == 0:
        ret = s_sep[0:0]

    else:

        abso = False
        if len(args) != 0 and len(args[0]) != 0:
            abso = args[0][0] == s_sep

        ret_l = []

        for i in args:

            if isinstance(i, list):

                joined = join(*i, must_be=must_be)

                ret_l += joined.split(s_sep)

            else:

                ret_l += i.split(s_sep)

        if len(ret_l) != 0:

            # NOTE: I know what this removes all empty lines from list,
            #       including the last ones, meaning adding slash to string.
            #       This is difference to os.path.join() and is intensional.

            while '' in ret_l:
                ret_l.remove('')

            while b'' in ret_l:
                ret_l.remove(b'')

        ret = s_sep.join(ret_l)

        if abso:
            ret = s_sep + ret

    return ret


def split(path):

    if not isinstance(path, (str, bytes)):
        raise ValueError("path must be str or bytes")

    s_sep = select_s_sep(path)

    absp = path.startswith(s_sep)

    path = _remove_double_sep(path)
    path = _remove_trailing_slash(path)

    ret = path.split(s_sep)

    while '' in ret:
        ret.remove('')

    while b'' in ret:
        ret.remove(b'')

    if absp:
        ret.insert(0, s_sep)

    return ret


def normpath(path):
    if not isinstance(path, (str, bytes)):
        raise ValueError("path must be str or bytes")
    return _remove_double_sep(os.path.normpath(path))


def abspath(path):
    if not isinstance(path, (str, bytes)):
        raise ValueError("path must be str or bytes")
    return _remove_double_sep(os.path.abspath(path))


def relpath(path, start):
    if not isinstance(path, (str, bytes)):
        raise ValueError("path must be str or bytes")
    if not isinstance(start, (str, bytes)):
        raise ValueError("start must be str or bytes")
    return _remove_double_sep(os.path.relpath(path, start))


def realpath(path):
    if not isinstance(path, (str, bytes)):
        raise ValueError("path must be str or bytes")
    return _remove_double_sep(os.path.realpath(path))


def abspaths(lst):

    ret = list()

    for i in lst:
        ret.append(abspath(i))

    return ret


def realpaths(lst):

    ret = list()

    for i in lst:
        ret.append(realpath(i))

    return ret


# NOTE: does not work
# def eval_abs_paths(lst, g, l):
#
#    """
#    Ensure(make) listed variables are(be) absolute path
#    """
#
#    for i in lst:
#        if i in l:
#            l[i] = abspath(l[i])
#
#    return


def prepend_path(lst, base):
    """
    Removes any trailing sep from base, and inserts it in the start of every
    lst item. if item not starts with separator, inserts it between base and
    item
    """

    lst = copy.copy(lst)

    while base.endswith(S_SEP):
        base = base[:-1]

    for i in range(len(lst)):
        sep = ''

        if lst[i][0] != S_SEP:
            sep = S_SEP

        lst[i] = base + sep + lst[i]

    return lst


def unprepend_path(lst, base):
    """
    Removes any trailing sep from base, and removes it from the start of every
    lst item.
    """

    if not isinstance(lst, list):
        raise TypeError("lst must be list")

    while base.endswith(S_SEP):
        base = base[:-1]

    for i in lst:
        if not (i + S_SEP).startswith(base + S_SEP):
            raise ValueError(
                "Not all items in lst have base `{}'".format(base)
                )

    lst = copy.copy(lst)

    base_l = len(base)

    for i in range(len(lst)):

        lst[i] = lst[i][base_l:]

    return lst


def insert_base(path, base):
    if not isinstance(path, str):
        raise ValueError("path must be str")
    return prepend_path([path], base)[0]


def remove_base(path, base):
    if not isinstance(path, str):
        raise ValueError("path must be str")
    return unprepend_path([path], base)[0]


def bases(lst):
    """
    Removes dirnames from paths
    """

    if not isinstance(lst, list):
        raise TypeError("lst must be list")

    ret = []

    for i in lst:
        ret.append(os.path.basename(i))

    return ret


def exclude_files_not_in_dirs(files, dirs):

    if not isinstance(files, list):
        raise TypeError("files must be list")

    if not isinstance(dirs, list):
        raise TypeError("dirs must be list")

    ret = []

    for i in files:

        d = os.path.dirname(i)

        if d in dirs:
            ret.append(i)

    return ret


def path_length(string):
    return len(split(string))


def _is_subpath_range_error(p1, p2):
    error = False
    for i in range(len(p1)):
        # TODO: check this
        if len(p1) < i + 1:
            error = True
            break

        # TODO: check this
        if len(p2) < i + 1:
            error = True
            break

        if p1[i] != p2[i]:
            error = True
            break

    return error


def get_subpath(basepath, fullpath):

    if not isinstance(basepath, str):
        raise ValueError("path must be str")

    if not isinstance(fullpath, str):
        raise ValueError("path_to_check must be str")

    p1 = split(abspath(basepath))
    p2 = split(abspath(fullpath))

    ret = None

    error = _is_subpath_range_error(p1, p2)

    if not error:
        ret = join(p2[len(p1):])

    return ret


def is_subpath(path_to_check, path):

    if not isinstance(path, str):
        raise ValueError("path must be str")

    if not isinstance(path_to_check, str):
        raise ValueError("path_to_check must be str")

    p1 = split(abspath(path))
    p2 = split(abspath(path_to_check))

    error = _is_subpath_range_error(p1, p2, i)

    ret = not error

    return ret


def is_subpath_real(path_to_check, path):

    if not isinstance(path, str):
        raise ValueError("path must be str")

    if not isinstance(path_to_check, str):
        raise ValueError("path_to_check must be str")

    p1 = split(realpath(path))
    p2 = split(realpath(path_to_check))

    error = _is_subpath_range_error(p1, p2)

    ret = not error

    return ret


def select_by_prefered_extension(lst, ext_lst):

    ret = None

    if len(lst) > 0:
        found = None
        for i in ext_lst:
            for j in lst:
                if j.endswith(i):
                    found = j
                    break
            if found is not None:
                break

        if found:
            ret = found
        else:
            ret = lst[0]

    return ret


def file_paths(filename):

    abs_full = abspath(filename)
    abs_dir = os.path.dirname(abs_full)
    abs_dir_real = realpath(abs_dir)
    abs_base = os.path.basename(abs_full)
    abs_dir_real_full = join(abs_dir_real, abs_base)

    real_full = realpath(abs_full)
    real_dir = os.path.dirname(real_full)
    real_base = os.path.basename(real_full)

    ret = {
        'abs_full': abs_full,
        'abs_dir': abs_dir,
        'abs_base': abs_base,

        'abs_dir_real': abs_dir_real,
        'abs_dir_real_full': abs_dir_real_full,

        'real_full': real_full,
        'real_dir': real_dir,
        'real_base': real_base
        }

    return ret




import threading
import time


class LoopedTimer:
    """
    Call for action every interval
    """

    def __init__(self, interval, action, args=None, kwargs=None):

        if args is None:
            args = tuple()

        if kwargs is None:
            kwargs = dict()

        self._interval = interval
        self._action = action
        self._args = args
        self._kwargs = kwargs

        self._thread = None

        self._stop_event = threading.Event()
        self._stopped_event = threading.Event()

        return

    def start(self):
        if not self._thread:
            self._stop_event.clear()
            self._stopped_event.clear()
            self._thread = threading.Thread(
                target=self._thread_target
                )
            self._thread.start()
        return

    def stop(self):
        self._stop_event.set()
        self._stopped_event.wait()
        return

    def _thread_target(self):

        while True:
            if self._stop_event.is_set():
                break

            time.sleep(self._interval)

            if self._stop_event.is_set():
                break

            self._action(*self._args, **self._kwargs)

        self._thread = None

        self._stopped_event.set()

        return




import os.path
import re
import subprocess

import wayround_org.utils.path


class SystemTypeInvalidFullName(Exception):
    pass


class SystemType:

    def __init__(
            self, fullname=None, cpu='i486', company='pc', kernel='linux', 
            os='gnu'
            ):

        self.fullname = fullname

        if isinstance(self.fullname, str):

            self._sane(self.fullname)

        else:
            self.cpu = cpu
            self.company = company
            self.os = os
            self.kernel = kernel

            self._sane(
                format_triplet(
                    cpu=cpu, company=company, kernel=kernel, os=os
                    )
                )

        return

    def _sane(self, s):

        if not isinstance(s, str):
            raise SystemTypeInvalidFullName(
                "Not valid fullname type: `{}'({})".format(s, type(s))
                )

        res = parse_triplet(s)

        if not res:
            raise SystemTypeInvalidFullName(
                "Not valid fullname: `{}'({})".format(s, type(s))
                )
        else:
            self.cpu = res[0]
            self.company = res[1]
            self.kernel = res[2]
            self.os = res[3]
            self.fullname = res[4]

        return

    def __str__(self):
        return self.fullname


def parse_triplet(str1):
    """
    Parse constitution triplet (``(.*?)-(.*?)-(.*)``), and return 3-tuple

    cpu-company-system

    where system can have one of these forms:

     os
     kernel-os

    """

    ret = None

    # FIXME: fix this function
    _debug=False

    if isinstance(str1, str):

        '''

        wd = os.path.dirname(wayround_org.utils.path.abspath(__file__))

        jd = os.path.join(wd, 'config.sub')

        if not os.path.isfile(jd):
            raise Exception("file not found: `{}'".format(jd))

        cmd = ['bash',
             jd,
             str1
             ]
        if _debug:
            print("cmd: {}".format(cmd))

        
        p = subprocess.Popen(
            cmd,
            cwd=wd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL
            )

        res = p.wait()

        if res != 0:
            if _debug:
                print("config.sub returned: {}".format(res))
            ret = None
            
        else:
        '''
        if True:
            '''
            com_res = p.communicate()

            out = str(com_res[0].splitlines()[0].strip(), 'utf-8')

            '''
            out = str1

            a = re.match(r'(?P<cpu>.*?)-(?P<company>.*?)-(?P<system>.*)', out)
            if not a:
                if _debug:
                    print("re not match")
            else:

                system = a.group('system')

                b = re.match(r'((?P<kernel>.*?)-)?(?P<os>.*)', system)

                if b:

                    ret = (
                        a.group('cpu'),
                        a.group('company'),
                        b.group('kernel'),
                        b.group('os'),
                        out
                        )

    return ret


def format_triplet(cpu='i486', company='pc', kernel='linux', os='gnu'):

    system = os

    if kernel:
        system = kernel + '-' + system

    return '{cpu}-{company}-{system}'.format(
        cpu=cpu,
        company=company,
        system=system
        )




"""
This module is for XHTML rendering

Some sort of template engine
"""

import logging
import re
import sys
import urllib.parse
import xml.sax.saxutils

import wayround_org.utils.dict
import wayround_org.utils.error
import wayround_org.utils.text


def pi(content):
    return {
        'type': 'pi',
        'content': content
        }


def comment(text):
    return {
        'type': 'comment',
        'content': text
        }


def dtd(text):
    return {
        'type': 'dtd',
        'content': text
        }


def cdata(text):
    return {
        'type': 'cdata',
        'content': text
        }


def char(text):
    return {
        'type': 'char',
        'content': text
        }


def static(text):
    return {
        'type': 'static',
        'content': text
        }


def html_head(
        title='',
        description='',
        keywords=[]
        ):
    return tag(
        'head',
        content={
            '00010_title': tag(
                'title',
                content=title
                ),
            '00020_description': tag(
                'meta',
                closed=True,
                attributes={
                    'name': 'description',
                    'content': description
                    }
                ),
            '00030_keywords': tag(
                'meta',
                closed=True,
                attributes={
                    'name': 'keywords',
                    'content': ' '.join(list(keywords))
                    }
                ),
            }
        )


def xhtml11(
        head=None,
        content=None,
        body_module=None,
        body_uid=None,
        body_js=None,
        body_css=None
        ):

    return {
        '00010_xml_pi': pi('xml version="1.0" encoding="UTF-8"'),
        '00015_html_dtd': dtd(
            'html PUBLIC "-//W3C//DTD XHTML 1.1//EN" '
            '"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd"'
            ),
        '00020_html': tag(
            'html',
            attributes={
                'version': '-//W3C//DTD XHTML 1.1//EN',
                'xmlns': 'http://www.w3.org/1999/xhtml',
                'xml:lang': 'en',
                'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
                'xsi:schemaLocation':
                    'http://www.w3.org/1999/xhtml '
                    'http://www.w3.org/MarkUp/SCHEMA/xhtml11.xsd'
                },
            closed=False,
            module=None,
            uid=None,
            required_css=[],
            required_js=[],
            content={
                '00010_head': head,
                '00020_body': tag(
                    'body',
                    module=body_module,
                    uid=body_uid,
                    required_css=body_css,
                    required_js=body_js,
                    content=content
                    )
                }
            )
        }


def html5(
        head=None,
        content=None,
        body_module=None,
        body_uid=None,
        body_js=None,
        body_css=None
        ):

    return {
        '00010_xml_pi': pi('xml version="1.0" encoding="UTF-8"'),
        '00015_html_dtd': dtd(
            'DOCTYPE xhtml'
            ),
        '00020_html': tag(
            'html',
            attributes={
                'lang': 'en'
                },
            closed=False,
            module=None,
            uid=None,
            required_css=[],
            required_js=[],
            content={
                '00010_head': head,
                '00020_body': tag(
                    'body',
                    module=body_module,
                    uid=body_uid,
                    required_css=body_css,
                    required_js=body_js,
                    content=content
                    )
                }
            )
        }


def tag(
        name,
        attributes=None,
        closed=False,
        module=None,
        uid=None,
        required_css=None,
        required_js=None,
        content=None,
        new_line_before_start=None,
        new_line_before_content=None,
        new_line_after_content=None,
        new_line_after_end=None
        ):
    ret = {
        'type': 'tag',
        'tag_info': {
            'name': name,
            'closed': closed,
            'attributes': attributes
            },
        'module': module,
        'uid': uid,
        'required_css': required_css,
        'required_js': required_js,
        'content': content
        }
    for i in [
            'new_line_before_start',
            'new_line_before_content',
            'new_line_after_content',
            'new_line_after_end'
            ]:
        if eval(i) is not None:
            ret[i] = eval(i)

    return ret


def check_unit(indict, path=[]):

    if isinstance(indict, str):
        return

    for i in list(indict.keys()):
        if not i in [
                'type',
                'tag_info',
                'content',
                'module',
                'uid',
                'required_css',
                'required_js',
                'new_line_before_start',
                'new_line_before_content',
                'new_line_after_content',
                'new_line_after_end'
                ]:
            raise KeyError("Wrong key `{}' at `{}'".format(i, '/'.join(path)))

    logging.debug("check_unit path {}".format('/'.join(path)))

    if len(path) >= 255:
        raise DictatorshipUnitTooDeep(
            "Dictatorship tree recursion limit reached `{}'".format(
                '/'.join(path)
                )
            )

    # Supplied data definitely must be a dict or list, otherwise - error
    if not isinstance(indict, (dict, list, str)):
        raise ValueError("Supplied data is not a dict or a list")

    # 'type' must be supplied
    if not 'type' in indict:
        raise MissingDictatorshipUnitAttribute(
            "Dictatorship unit type missing"
            )

    # 'type' must be one of following
    if not indict['type'] in [
            'tag', 'dtd', 'comment', 'cdata', 'static',
            'pi', 'char'
            ]:
        raise ValueError("Wrong dictatorship unit type")

    # If 'type' is 'tag', then check 'tag_info' and everything,
    # what underlie
    if indict['type'] == 'tag':
        if not 'tag_info' in indict \
                or not isinstance(indict['tag_info'], dict):
            raise MissingDictatorshipUnitAttribute(
                "Dictatorship unit type is `tag', but not `tag_info' supplied"
                )

        else:
            for i in list(indict['tag_info'].keys()):
                if not i in [
                        'attributes',
                        'name',
                        'closed'
                        ]:
                    raise KeyError(
                        "Wrong key `{}' at `{}'['tag_info']".format(
                            i, '/'.join(path)
                            )
                        )

            # tag name MUST be supplied and must be a string!
            if not 'name' in indict['tag_info']:
                raise MissingDictatorshipUnitAttribute(
                    "`name' not supplied in dictatorship unit `tag_info'"
                    )
            else:
                if not isinstance(indict['tag_info']['name'], str):
                    raise TypeError("tag `name' must be a string")

            # attributes CAN be supplied or CAN be a None or a dict
            if 'attributes' in indict['tag_info']:
                if indict['tag_info']['attributes'] is None:
                    indict['tag_info']['attributes'] = {}
                else:
                    if not isinstance(indict['tag_info']['attributes'], dict):
                        raise TypeError("tag `attributes' must be dict")
                    else:
                        # attribute values can be a strings or callabels
                        for i in list(indict['tag_info']['attributes'].keys()):
                            if not isinstance(
                                    indict['tag_info']['attributes'][i], str
                                    ):
                                raise TypeError(
                                    "tag `attributes' dict "
                                    "values must be strings"
                                    )
            else:
                indict['tag_info']['attributes'] = {}

            # 'tag_info' 'closed' attribute CAN be supplied and CAN be
            #  None or bool.
            if 'closed' in indict['tag_info']:
                if not isinstance(indict['tag_info']['closed'], bool):
                    raise TypeError("tag `closed' attribute can be only bool")
            else:
                indict['tag_info']['closed'] = False

    for i in ['required_css', 'required_js']:
        if i in indict:
            if indict[i] is None:
                indict[i] = []
            elif not isinstance(indict[i], list):
                raise TypeError("`{}' can only be list or None".format(i))
            else:
                for j in indict[i]:
                    if not isinstance(j, str):
                        raise TypeError(
                            "All `{}' values must be strings".format(i)
                            )

        else:
            indict[i] = []

    for i in ['module', 'uid']:
        if i in indict:
            if indict[i] is not None:
                if isinstance(indict[i], str):
                    if not re.match(r'[a-zA-Z][a-zA-Z\-]*', indict[i]):
                        raise ValueError(
                            "Wrong `{i}' value at `{path}'".format_map(
                                {
                                    'path': '/'.join(path),
                                    'i': i
                                    }
                                )
                            )
                else:
                    raise ValueError("`{}' can be None or str".format(i))

    if not 'content' in indict:
        indict['content'] = ''
    elif indict['content'] is None:
        indict['content'] = ''
    elif isinstance(indict['content'], str):
        pass
    elif isinstance(indict['content'], dict):
        pass
    elif isinstance(indict['content'], list):
        pass
    else:
        raise ValueError(
            "wrong unit content value: `{}', but "
            "must be None, str, dict or list".format(
                type(indict['content'])
                )
            )

    default_new_line_before_start = False
    default_new_line_before_content = False
    default_new_line_after_content = False
    default_new_line_after_end = False

    if indict['type'] == 'tag':
        default_new_line_before_start = True
        if indict['tag_info']['closed']:
            default_new_line_after_content = False
        else:
            if isinstance(indict['content'], (dict, list)):
                default_new_line_after_content = True

    elif indict['type'] in ['tag', 'comment', 'dtd', 'pi']:
        default_new_line_after_end = True

    for i in [
            ('new_line_before_start', default_new_line_before_start),
            ('new_line_before_content', default_new_line_before_content),
            ('new_line_after_content', default_new_line_after_content),
            ('new_line_after_end', default_new_line_after_end)
            ]:
        if i[0] in indict:
            if not isinstance(indict[i[0]], bool):
                raise TypeError("Wrong `{}' value type".format(i[0]))
        else:
            indict[i[0]] = i[1]

    return


def render_attributes(indict, path=[], tagname='', xml_indent_size=2):
    ret = ''

    inaddr_l = len(path)

    indent = wayround_org.utils.text.fill(' ', inaddr_l * xml_indent_size)
    nameindent = wayround_org.utils.text.fill(' ', len(tagname))

    attrs = []

    keys = sorted(indict.keys())

    for i in keys:

        if ret != '':
            ret += ' '

        value = ''
        if isinstance(indict[i], str):
            value = indict[i]
        else:
            raise ValueError(
                "One of attribute values (`{}') is not str".format(i)
                )

        if isinstance(ret, str):

            try:
                attrs.append(
                    '{name}={value}'.format_map(
                        {
                            'name': i,
                            'value': xml.sax.saxutils.quoteattr(value)
                            }
                        )
                    )
            except:
                ret = 1
                break

    if isinstance(ret, str):
        ind_req = False
        for i in attrs:
            if len(i) > 80:
                ind_req = True
                break

        first = True

        curr_attr_i = 0
        attrs_l = len(attrs)

        for i in attrs:
            ind = ''
            if ind_req and not first:
                ind = "\n{indent} {nameindent} ".format_map(
                    {
                        'indent': indent,
                        'nameindent': nameindent
                        }
                    )

            ret += "{ind}{new_attr}".format_map(
                {
                    'ind': ind,
                    'new_attr': i
                    }
                )

            if curr_attr_i < attrs_l - 1:
                ret += ' '

            if first:
                first = False

            curr_attr_i += 1

    return ret


def dict_tree_to_xml_render(tree):
    b = DictTreeToXMLRenderer(2, True, True)

    b.set_tree(tree)

    ret = b.render()

    b.print_log()

    return ret


class DictatorshipUnitTooDeep(Exception):
    pass


class MissingDictatorshipUnitAttribute(Exception):
    pass


class ModulesDirError(Exception):
    pass


class DictTreeToXMLRenderer:

    """
    renderer for dict structures described in this modules

    use settings on init to adjust needed parameters, set_tree() method to
    set required tree and then user method render() to get the string
    """

    def __init__(
            self,
            xml_indent_size=2,
            generate_css=False,
            generate_js=False,
            log_size=100,
            space_before_closing_slash=False,
            css_and_js_holder=None
            ):

        # here linedup units are listed. key is path
        self.units = {}

        # here will be stored
        self.tree_dict = {}

        # those four attributes are for code formatting
        # purposes
        self.xml_indent_size = xml_indent_size
        self.xml_indent = wayround_org.utils.text.fill(' ', xml_indent_size)

        self.css_and_js_holder = css_and_js_holder

        self.css_placeables = []
        self.js_placeables = []

        self.log = []

        self.log_size = log_size

        self.generate_css = generate_css
        self.generate_js = generate_js
        self.space_before_closing_slash = space_before_closing_slash

        return

    def do_log(self, text):

        log_l = len(self.log)

        if log_l >= self.log_size:
            self.log = self.log[-self.log_size:]

        self.log.append(text)

    def print_log(self):
        for i in self.log:
            print(i)

    def set_tree(self, indict):
        self.tree_dict = indict

    def _lineup_tree(self, indict, already_added, path=[]):

        logging.debug("_lineup_tree path {}".format('/'.join(path)))

        ret = 0

        if not isinstance(indict, (dict, list)):
            raise TypeError("This method accepts only dict or list")

        if self.check_range(indict) != 0:
            self.do_log(
                "_lineup_tree wrong input data at path `{}'".format(
                    '/'.join(path)
                    )
                )

            ret = 1
        else:

            index = -1

            if isinstance(indict, dict):
                keys = sorted(indict.keys())
            else:
                keys = indict

            for i in keys:

                index += 1

                if isinstance(indict, dict):
                    unit = indict[i]
                elif isinstance(indict, list):
                    unit = i

                if isinstance(indict, dict):
                    path_name = i
                elif isinstance(indict, list):
                    path_name = str(index)

                if not id(unit) in already_added:

                    try:
                        tmp = '/'.join(path + [path_name])
                    except:
                        raise

                    if not tmp in self.units:
                        self.units[tmp] = unit

                    del(tmp)

                    already_added.add(id(unit))

                if 'content' in unit:
                    if isinstance(unit['content'], (dict, list)):

                        if self._lineup_tree(
                                unit['content'],
                                already_added,
                                path=path + [path_name]
                                ) != 0:
                            ret = 2

        return ret

    def lineup_tree(self):
        self.units = {}
        already_added = set()
        ret = self._lineup_tree(
            self.tree_dict, already_added, path=[]
            )

        return ret

    def check_tree(self):
        ret = 0

        keys = sorted(self.units.keys())

        for i in keys:
            try:
                check_unit(
                    self.units[i], i.split('/')
                    )
            except:
                self.do_log(
                    "-e- Error while checking `{path}'\n{exc_info}".format_map(
                        {
                            'path': i,
                            'exc_info':
                                wayround_org.utils.error.return_exception_info(
                                    sys.exc_info(),
                                    tb=True
                                    )
                            }
                        )
                    )
                ret = 1
                break

        return ret

    def find_required_css_and_js(self):

        self.css_placeables = []
        self.js_placeables = []

        keys = sorted(self.units.keys())

        for i in keys:

            if not self.generate_css and not self.generate_js:
                break

            if not 'module' in self.units[i] \
                    or not 'uid' in self.units[i]:
                continue

            if self.generate_css:
                if 'required_css' in self.units[i]:
                    for j in self.units[i]['required_css']:
                        placeable_name = \
                            "{module}/{uid}/{required}".format_map(
                                {
                                    'module': self.units[i]['module'],
                                    'uid': self.units[i]['uid'],
                                    'required': j,
                                    }
                                )
                        if not placeable_name in self.css_placeables:
                            self.css_placeables.append(placeable_name)

            if self.generate_js:
                if 'required_js' in self.units[i]:
                    for j in self.units[i]['required_js']:
                        placeable_name = \
                            "{module}/{uid}/{required}".format_map(
                                {
                                    'module': self.units[i]['module'],
                                    'uid': self.units[i]['uid'],
                                    'required': j,
                                    }
                                )
                        if not placeable_name in self.js_placeables:
                            self.js_placeables.append(placeable_name)

        return

    def _place_found_css_or_js(
            self,
            i,
            placement_i,
            typ,
            css_path_renderer=css_path_renderer,
            js_path_renderer=js_path_renderer
            ):

        if not typ in ['css', 'js']:
            raise ValueError("Wrong `typ' parameter value")

        if typ == 'css':
            new_val = tag(
                'link',
                closed=True,
                attributes={
                    'type': 'text/css',
                    'href': css_path_renderer(i),
                    'rel': 'stylesheet'
                    }
                )

        elif typ == 'js':
            new_val = tag(
                'script',
                attributes={
                    'type': 'text/javascript',
                    'src': js_path_renderer(i),
                    }
                )
        else:
            raise Exception("Wrong programming")

        check_unit(new_val)

        if isinstance(self.css_and_js_holder['content'], dict):
            wayround_org.utils.dict.append(
                self.css_and_js_holder['content'],
                new_val
                )

        elif isinstance(self.css_and_js_holder['content'], list):
            self.css_and_js_holder['content'].append(new_val)

    def place_found_css_and_js(
            self,
            css_path_renderer=css_path_renderer,
            js_path_renderer=js_path_renderer
            ):

        if not isinstance(self.css_and_js_holder, dict):
            raise ValueError("Wrong self.css_and_js_holder")

        if not isinstance(self.css_and_js_holder['content'], (dict, list)):
            raise TypeError(
                "self.css_and_js_holder['content'] can be only list or dict"
                )

        placement_i = 0

        if self.generate_css:
            for i in self.css_placeables:

                self._place_found_css_or_js(
                    i,
                    placement_i,
                    'css',
                    css_path_renderer=css_path_renderer,
                    js_path_renderer=js_path_renderer
                    )

                placement_i += 1

        if self.generate_js:
            for i in self.js_placeables:

                self._place_found_css_or_js(
                    i,
                    placement_i,
                    'js',
                    css_path_renderer=css_path_renderer,
                    js_path_renderer=js_path_renderer
                    )

                placement_i += 1

        return

    def _render_unit(self, root, unit, path, path_name, indent):

        if isinstance(unit, str):
            return xml.sax.saxutils.escape(unit)

        new_line_before_start = ''
        if unit['new_line_before_start']:
            new_line_before_start = '\n{}'.format(indent)

        new_line_before_content = ''
        if unit['new_line_before_content']:
            new_line_before_content = '\n'

        new_line_after_content = ''
        if unit['new_line_after_content']:
            new_line_after_content = '\n{}'.format(indent)

        new_line_after_end = ''
        if unit['new_line_after_end']:
            new_line_after_end = '\n'

        start = ''
        content = ''
        end = ''

        if unit['type'] == 'comment':
            start = '<!-- '

            content = str(unit['content'])

            content = content.replace('--', '-')

            end = ' -->'

        elif unit['type'] == 'pi':
            start = '<?'
            content = str(unit['content'])
            end = '?>'

        elif unit['type'] == 'dtd':
            start = '<!DOCTYPE '
            content = str(unit['content'])
            end = '>'

        elif unit['type'] == 'cdata':
            start = '<![CDATA['
            content = str(unit['content']).replace(']]>', '')
            end = ']]>'

        elif unit['type'] == 'char':
            start = ''
            content = xml.sax.saxutils.escape(str(unit['content']))
            end = ''

        elif unit['type'] == 'static':
            start = ''
            content = str(unit['content'])
            end = ''

        elif unit['type'] == 'tag':

            if (('module' in unit)
                    and ('uid' in unit)
                    and (isinstance(unit['module'], str))
                    and (isinstance(unit['uid'], str))
                ):
                class_list = list()

                if (
                        ('class' in unit['tag_info']['attributes'])
                        and
                        (isinstance(
                            unit['tag_info']['attributes']['class'], str))
                        ):
                    class_list += \
                        (unit['tag_info']['attributes']['class']).split(' ')

                class_list.append(unit['module'] + '---' + unit['uid'])

                class_list = sorted(set(class_list))

                unit['tag_info']['attributes']['class'] = ' '.join(class_list)

                class_list = []

            attributes = ''
            if unit['tag_info']['attributes'] is not None:
                attributes = render_attributes(
                    unit['tag_info']['attributes'],
                    path,
                    tagname=unit['tag_info']['name'],
                    xml_indent_size=self.xml_indent_size
                    )

            closing_slash = ''
            if unit['tag_info']['closed']:
                closing_slash = '/'

            space_before_attributes = ''
            if attributes != '':
                space_before_attributes = ' '

            space_before_closing_slash = ''
            if closing_slash != '':
                if self.space_before_closing_slash:
                    space_before_closing_slash = ' '
                else:
                    space_before_closing_slash = ''

            start = (
                '<{tagname}{space_before_attributes}{attributes}'
                '{space_before_closing_slash}{closing_slash}>'
                ).format_map(
                    {
                        'tagname': unit['tag_info']['name'],
                        'space_before_attributes': space_before_attributes,
                        'attributes': attributes,
                        'space_before_closing_slash':
                        space_before_closing_slash,
                        'closing_slash': closing_slash
                        }
                    )

            if not unit['tag_info']['closed']:

                if isinstance(unit['content'], str):
                    content = xml.sax.saxutils.escape(unit['content'])
                elif isinstance(unit['content'], (dict, list)):

                    content = self._render(
                        root, unit['content'], path=path + [path_name]
                        )
                elif unit['content'] is None:
                    content = ''

                else:
                    content = str(unit['content'])

                end = '</{}>'.format(unit['tag_info']['name'])
            else:
                content = ''
                end = ''

        else:
            raise ValueError("Wrong type at `{}'".format('/'.join(path)))

        ret = (""
               + "{new_line_before_start}"
               + "{start}"
               + "{new_line_before_content}"
               + "{content}"
               + "{new_line_after_content}"
               + "{end}"
               + "{new_line_after_end}").format_map(
            {
                'new_line_before_start': new_line_before_start,
                'new_line_before_content': new_line_before_content,
                'new_line_after_content': new_line_after_content,
                'new_line_after_end': new_line_after_end,
                'start': start,
                'content': content,
                'end': end
                }
            )

        return ret

    def _render(self, root, indict, path=[]):

        ret = ''

        inaddr_l = len(path)

        indent = wayround_org.utils.text.fill(
            ' ',
            inaddr_l * self.xml_indent_size
            )

        index = -1

        if isinstance(indict, dict):
            keys = sorted(indict.keys())
        elif isinstance(indict, list):
            keys = indict
        else:
            raise TypeError("This method accepts only dict or list")

        for i in keys:
            index += 1

            unit = None

            if isinstance(indict, dict):
                unit = indict[i]
            elif isinstance(indict, list):
                unit = i
            else:
                raise TypeError("This method accepts only dict or list")

            if isinstance(indict, dict):
                path_name = i
            elif isinstance(indict, list):
                path_name = str(index)

            ret += self._render_unit(root, unit, path, path_name, indent)

        return ret

    def render(
            self,
            css_path_renderer=css_path_renderer,
            js_path_renderer=js_path_renderer
            ):

        ret = ''

        if self.lineup_tree() != 0:
            self.do_log("-e- Some errors liningup dict tree")
            ret = 1

        if isinstance(ret, str):
            if self.check_tree() != 0:
                ret = 2

        if isinstance(ret, str):

            if self.css_and_js_holder:
                self.find_required_css_and_js()
                self.place_found_css_and_js(
                    css_path_renderer=css_path_renderer,
                    js_path_renderer=js_path_renderer
                    )

        if isinstance(ret, str):
            ret = self._render(self.tree_dict, self.tree_dict, path=[])

        return ret

    def check_range(self, indict):
        ret = 0

        if not isinstance(indict, (dict, list)):
            self.do_log(
                "-e- Supplied data is not a dict and not list: {}".format(
                    repr(indict)
                    )
                )
            ret = 1

        else:

            for i in indict:

                if isinstance(indict, dict):
                    unit = indict[i]
                elif isinstance(indict, list):
                    unit = i
                else:
                    raise TypeError("This method accepts only dict or list")

                if not isinstance(unit, (dict, str)):
                    self.do_log(
                        "-e- Dictatorship `{}' element value "
                        "is not a dict and not str".format(str(unit))
                        )

                    ret = 2
                    break

        return ret


def css_path_renderer(inname):
    module, uid, file = inname.split('/')[0:3]

    return "css?module={module}&uid={uid}&file={file}".format_map(
        {
            'module': urllib.parse.quote(
                module, encoding='utf-8', errors='strict'
                ),
            'uid': urllib.parse.quote(
                uid, encoding='utf-8', errors='strict'
                ),
            'file': urllib.parse.quote(
                file, encoding='utf-8', errors='strict'
                )
            }
        )


def js_path_renderer(inname):
    module, uid, file = inname.split('/')[0:3]

    return "js?module={module}&uid={uid}&file={file}".format_map(
        {
            'module': urllib.parse.quote(
                module, encoding='utf-8', errors='strict'
                ),
            'uid': urllib.parse.quote(
                uid, encoding='utf-8', errors='strict'
                ),
            'file': urllib.parse.quote(
                file, encoding='utf-8', errors='strict'
                )
            }
        )


def test():
    # Upper dict element with names (like following) is called
    # `dictatorship range'.
    # Dictatorship range's keys must point only on dict-s,
    # otherwise it is an error.

    logging.basicConfig(level='DEBUG')

    a = {
        '000_xml_pi': {
            'type': 'pi',
            'content': 'xml version="1.1" encoding="UTF-8"',
            },
        '010_xhtml_doctype': {
            'type': 'dtd',
            'content':
                'html PUBLIC "-//W3C//DTD XHTML 1.1//EN" '
                '"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd"'
            },
        '100_html_tag': {
            # Dictatorship gange pointet dicts (like this)
            # is called `dictatorship unit'

            # Any dict can be one of folloving types:
            # 'tag', 'dtd', 'comment', 'cdata', 'static', 'char'
            # 'pi' (Processing instructions).
            # If type is 'tag',then 'tag_info' subdict is required
            'type': 'tag',

            # inserts new lines
            # 'new_line_before_start': False,
            # 'new_line_before_content': False,
            # 'new_line_after_content': False,
            # 'new_line_after_end': False,

            # 'tag_info' required to present if type is 'tag'.
            # 'tag_info' must be a dict.
            'tag_info': {
                # any valid tag name
                'name': 'html',

                # None or dict
                # dict values must be strings
                'attributes': {
                    'version': '-//W3C//DTD XHTML 1.1//EN',
                    'xmlns': 'http://www.w3.org/1999/xhtml',
                    'xml:lang': 'en',
                    'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
                    'xsi:schemaLocation':
                        'http://www.w3.org/1999/xhtml '
                        'http://www.w3.org/MarkUp/SCHEMA/xhtml11.xsd'
                    },

                # (bool) if True - no contents is
                # passible and any info supplied in 'contents'
                # will be omited. Default is False.
                'closed': False
                },

            # str, dict, None, callable
            # None - exchenged to empty string
            # string - used as is
            # dict - assumed tobe new
            # `dictatorship range'
            # callable - must return string
            'content': {
                '10_head': {
                    'type': 'tag',
                    'tag_info': {
                        'name': 'head'
                        },
                    'content': {
                        'title': {
                            'type': 'tag',
                            'tag_info': {
                                'name': 'title'
                                },
                            'content': 'Page Title'
                            }
                        }
                    },
                '20_body': {
                    'type': 'tag',
                    'tag_info': {
                        'name': 'body'
                        },
                    'module': 'core',
                    'uid': 'body',
                    'required_css': ['main.css', 'body.css'],
                    'required_js': ['main.js'],
                    }
                }

            }
        }

    print("Dictator test #1")
    b = DictTreeToXMLRenderer(2, True, True)
    b.set_tree(a)
    ret = b.render()
    print(repr(b.units))
    b.print_log()
    print(ret)
    return


def test2():
    logging.basicConfig(level='DEBUG')
    print("Dictator test #2")
    b = DictTreeToXMLRenderer(2, True, True)
    b.set_tree(html5())
    ret = b.render()
    print(repr(b.units))
    b.print_log()
    print(ret)
